# Copyright 2011-2014 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from fuselage import error, platform, provider, resources
from fuselage.changes import ShellCommand


class User(provider.Provider):

    policies = (resources.user.UserApplyPolicy,)

    def get_user_info(self):
        fields = ("name", "passwd", "uid", "gid", "gecos", "dir", "shell")

        username = self.resource.name

        try:
            info_tuple = platform.getpwnam(username)
        except KeyError:
            info = {f: None for f in fields}
            info["exists"] = False
            info["disabled-login"] = False
            info["disabled-password"] = False
            return info

        info = {
            "exists": True,
            "disabled-login": False,
            "disabled-password": False,
        }
        for i, field in enumerate(fields):
            info[field] = info_tuple[i]

        try:
            shadow = platform.getspnam(username)
            info["passwd"] = shadow.sp_pwd
            if shadow.sp_pwd == "!":
                info["disabled-login"] = True
        except KeyError:
            info["passwd"] = ""
            info["disabled-login"] = False

        return info

    def apply(self):
        info = self.get_user_info()

        if info["exists"]:
            command = ["usermod"]
            changed = False  # we may not change anything yet
        else:
            command = ["useradd", "-N"]
            changed = True  # we definitely make a change

        name = self.resource.name

        fullname = self.resource.fullname
        if fullname and info["gecos"] != fullname:
            command.extend(["--comment", self.resource.fullname])
            changed = True

        password = self.resource.password
        if password and not info["exists"]:
            command.extend(["--password", self.resource.password])
            changed = True

        home = self.resource.home
        if home and info["dir"] != home:
            command.extend(["--home", self.resource.home])
            command.append("-m")
            changed = True

        uid = self.resource.uid
        if uid and info["uid"] != int(uid):
            command.extend(["--uid", str(self.resource.uid)])
            changed = True

        gid = self.resource.gid
        group = self.resource.group
        if gid or group:
            if gid:
                gid = int(gid)
                if gid != info["gid"]:
                    command.extend(["--gid", str(self.resource.gid)])
                    changed = True
            else:
                try:
                    gid = platform.getgrnam(group).gr_gid
                except KeyError:
                    self.raise_or_log(
                        error.InvalidGroup('Group "%s" is not valid' % group)
                    )
                    gid = "GID_CURRENTLY_UNASSIGNED"

                if gid != info["gid"]:
                    command.extend(["--gid", str(gid)])
                    changed = True

        groups = self.resource.groups
        if groups:
            desired_groups = set(groups)
            current_groups = {
                g.gr_name for g in platform.getgrall() if name in g.gr_mem
            }

            append = self.resource.append
            if append and len(desired_groups - current_groups) > 0:
                if info["exists"]:
                    command.append("-a")
                command.extend(["-G", ",".join(desired_groups - current_groups)])
                changed = True
            elif not append and desired_groups != current_groups:
                command.extend(["-G", ",".join(desired_groups)])
                changed = True

        shell = self.resource.shell
        if shell and shell != info["shell"]:
            command.extend(["--shell", str(self.resource.shell)])
            changed = True

        disabled_login = self.resource.disabled_login
        if disabled_login and not info["disabled-login"]:
            command.extend(["--password", "!"])
            changed = True

        system = self.resource.system
        if not info["exists"] and system:
            command.extend(["--system"])
            changed = True

        command.extend([self.resource.name])

        if changed:
            try:
                self.change(ShellCommand(command))
            except error.SystemError as exc:
                raise error.UserAddError(
                    "useradd returned error code %d" % exc.returncode
                )
        return changed


class UserRemove(provider.Provider):

    policies = (resources.user.UserRemovePolicy,)

    def apply(self):
        try:
            platform.getpwnam(self.resource.name.encode("utf-8"))
        except KeyError:
            # If we get a key errror then there is no such user. This is good.
            return False

        command = ["userdel", self.resource.name]

        try:
            self.change(ShellCommand(command))
        except error.SystemError as exc:
            raise error.UserAddError(
                "Removing user %s failed with return code %d"
                % (self.resource, exc.returncode)
            )

        return True



# Copyright 2012-2014 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from fuselage import error, platform, provider, resources
from fuselage.changes import ShellCommand


class Mounted(provider.Provider):

    policies = (resources.checkout.CheckoutSyncPolicy,)

    @classmethod
    def isvalid(self, policy, resource):
        return resource.scm in ("dummy", "mounted", "mount")

    def apply(self):
        if not platform.exists(self.resource.name):
            return self.raise_or_log(error.PathComponentMissing(self.resource.name))

        for path in self.resource.changes:
            if platform.exists(path):
                self.logger.debug(f"Faking changes to {path}")
                self.change(ShellCommand(["touch", "-ac", path]))
            else:
                self.logger.debug(f"Not faking changes to {path} (file not present)")

        self.logger.debug("Faking that checkout changed")
        return True



# Copyright 2011-2014 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import shlex

from fuselage import error, platform
from fuselage.changes import base
from fuselage.utils import force_str


class ShellCommand(base.Change):

    """Execute and log a change"""

    changed = True

    def __init__(
        self,
        command,
        shell=None,
        stdin=None,
        cwd=None,
        env=None,
        user="root",
        group=None,
        umask=None,
        expected=0,
        logas=None,
    ):
        if isinstance(command, list):
            self.command = command
        elif isinstance(command, str):
            self.command = shlex.split(command)
        else:
            raise RuntimeError("'command' must be a list or string")

        # Commands can be logged differently to what is executed, so that passwords aren't revealed
        if logas:
            if isinstance(logas, str):
                self.logas = shlex.split(self.command)
            else:
                self.logas = logas
        else:
            self.logas = self.command

        self.stdin = stdin
        self.cwd = cwd

        self.env = {}
        if platform.platform == "win32":
            self.env.update({"PATH": os.environ.get("PATH", "")})
        else:
            self.env.update(
                {
                    "PATH": "/usr/local/sbin:/usr/local/bin:/usr/sbin:/usr/bin:/sbin:/bin",
                }
            )
        if env:
            self.env.update(env)

        self.user = user
        self.group = group
        self.umask = umask
        self.expected = expected

    def _tounicode(self, value):
        """Ensure all elements of the list are unicode"""

        def uni(x):
            if isinstance(x, str):
                return x
            return x.encode("utf-8")

        return list(map(uni, value))

    def command_exists(self, command):
        if command[0].startswith("./"):
            if len(command[0]) <= 2:
                return False
            return platform.exists(os.path.join(self.cwd, command[0][2:]))

        elif command[0].startswith("/"):
            return platform.exists(command[0])

        for path in self.env["PATH"].split(platform.pathsep):
            if platform.exists(os.path.join(path, command[0])):
                return True

    def apply(self, ctx):
        command, logas = self.command, self.logas

        ctx.changelog.critical("# " + " ".join([force_str(val) for val in logas]))

        if not self.command_exists(command):
            ctx.raise_or_log(error.BinaryMissing("Command '%s' not found" % command[0]))

        if platform.getpwnam and self.user:
            try:
                platform.getpwnam(self.user)
            except KeyError:
                ctx.raise_or_log(error.InvalidUser("User '%s' not found" % self.user))

        if platform.getgrnam and self.group:
            try:
                platform.getgrnam(self.group)
            except KeyError:
                ctx.raise_or_log(error.InvalidGroup("User '%s' not found" % self.group))

        if self.cwd:
            if not platform.isdir(self.cwd):
                ctx.raise_or_log(
                    error.PathComponentNotDirectory("%r not a directory" % self.cwd)
                )

        if ctx.simulate:
            self.returncode = 0
            self.stdout = ""
            self.stderr = ""
            return

        self.stdout, self.stderr = platform.check_call(
            command=command,
            user=self.user,
            group=self.group,
            umask=self.umask,
            env=self.env,
            cwd=self.cwd,
            expected=self.expected,
            logger=ctx.changelog,
        )



# Copyright 2011-2014 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from fuselage.argument import Boolean, FullPath, Integer, List, String
from fuselage.policy import NAND, Policy, Present
from fuselage.resource import Resource


class User(Resource):

    """A resource representing a UNIX user in the password database. The underlying implementation currently uses the "useradd" and "usermod" commands to implement this resource.

    This resource can be used to create, change or delete UNIX users.

    For example::

        User:
          name: django
          fullname: Django Software Owner
          home: /var/local/django
          system: true
          disabled-password: true

    """

    name = String()
    """ The username this resource represents. """

    password = String()
    """ The encrypted password, as returned by crypt(3). You should make sure
    this password respects the system's password policy. """

    fullname = String()
    """ The comment field for the password file - generally used for the user's full name. """

    home = FullPath()
    """ The full path to the user's home directory. """

    uid = Integer()
    """ The user identifier for the user. This must be a non-negative integer. """

    gid = Integer()
    """ The group identifier for the user. This must be a non-negative integer. """

    group = String()
    """ The primary group for the user, if you wish to specify it by name. """

    groups = List()
    """ A list of supplementary groups that the user should be a member of. """

    append = Boolean(default=True)
    """ A boolean that sets how to apply the groups a user is in. If true then yaybu will
    add the user to groups as needed but will not remove a user from a group. If false then yaybu will replace
    all groups the user is a member of. Thus if a process outside of yaybu adds you to a group,
    the next deployment would remove you again. """

    system = Boolean(default=True)
    """ A boolean representing whether this user is a system user or not. This only takes effect on
    creation - a user cannot be changed into a system user once created
    without deleting and recreating the user. """

    shell = FullPath(default="/bin/bash")
    """ The full path to the shell to use. """

    disabled_password = Boolean(default=False)
    """ A boolean for whether the password is locked for this account. """

    disabled_login = Boolean(default=False)
    """ A boolean for whether this entire account is locked or not. """


class UserApplyPolicy(Policy):

    """Create or change the specified user.

    It might not be possible to apply some changes to existing users, for example
    the systeam attribute only makes sense at the point a user is created.
    """

    resource = User
    name = "apply"
    default = True

    signature = (
        Present("name"),
        NAND(Present("password"), Present("disabled_login")),
    )


class UserRemovePolicy(Policy):

    """Remove an existing user. This is not recommended in general - you
    should lock existing accounts instead to preserve file ownership metadata
    information."""

    resource = User
    name = "remove"
    default = False



# Copyright 2013-2014 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from fuselage import error, platform, provider, resources
from fuselage.changes import EnsureDirectory, EnsureFile, ShellCommand
from fuselage.utils import force_bytes

try:
    from urllib.parse import quote, urlparse, urlunparse
except ImportError:
    from urllib import quote

    from urlparse import urlparse, urlunparse


hgrc = """
[paths]
default = %(repository)s
[extensions]
should = %(path)s/.hg/should.py
"""

mercurial_ext = b"""
from mercurial import util, hg, node

def should_pull(ui, repo, **opts):
    default_path = repo.ui.configlist('paths', 'default')[0]

    if not hasattr(hg, "peer"):
        source, revs, checkout = hg.parseurl(ui.expandpath(default_path), [])
        peer = hg.repository(ui, source)
    else:
        peer = hg.peer(ui, {}, default_path)

    remote_branches = peer.branchmap()
    local_branches = repo.branchmap()

    if opts['branch'] not in remote_branches:
        raise util.Abort('NO_SUCH_BRANCH: "%s" is not in the repository' % opts['branch'])

    if opts['branch'] not in local_branches:
        raise util.Abort('PULL: "%s" in not local, but is available in remote' % opts['branch'])

    if not opts['tag']:
        if remote_branches[opts['branch']] != local_branches[opts['branch']]:
            raise util.Abort('PULL: "%s" is out of date' % opts['branch'])

        ui.write("OK: Up to date")

    else:
        if not opts['tag'] in repo.tags().keys():
           raise util.Abort('PULL: "%s" is not in local' % opts['tag'])

        ui.write("OK: Tag is already available locally")


def should_update(ui, repo, **opts):
    if opts['tag']:
        target = opts['tag']
        revmap = repo.tags()
        if not target in revmap:
            raise util.Abort("FAIL: Tag '%s' not found locally" % target)
        targetrev = [revmap[target]]
    else:
        target = opts['branch']
        revmap = repo.branchmap()
        if not target in revmap:
            raise util.Abort("FAIL: Branch '%s' not found locally" % target)
        targetrev = revmap[target]

    localrev = [node.short(p.node()) for p in repo[None].parents()]
    targetrev = [node.short(p) for p in targetrev]

    if localrev != targetrev:
        raise util.Abort("UPDATE: Checkout is at '%s', target '%s' is at revision '%s'" % (localrev, target, targetrev))

    ui.write("OK: Checkout is up to date")


cmdtable = {
    'should-pull': (should_pull, [('b', 'branch', 'default', 'Branch to track'), ('t', 'tag', '', 'Tag to track')], '[options]'),
    'should-update': (should_update, [('b', 'branch', 'default', 'Branch to track'), ('t', 'tag', '', 'Tag to track')], '[options]'),
    }
"""


def _inject_credentials(url, username=None, password=None):
    if username and password:
        p = urlparse(url)
        netloc = "{}:{}@{}".format(quote(username, ""), quote(password, ""), p.hostname)
        if p.port:
            netloc += ":" + str(p.port)
        url = urlunparse((p.scheme, netloc, p.path, p.params, p.query, p.fragment))
    return url


class Mercurial(provider.Provider):

    policies = (resources.checkout.CheckoutSyncPolicy,)

    @classmethod
    def isvalid(self, policy, resource):
        return resource.scm and resource.scm.lower() == "mercurial"

    def get_hg_command(self, action, *args):
        command = [
            "hg",
            action,
        ]

        command.extend(list(args))
        return command

    def info(self, action, *args):
        return platform.check_call(
            command=self.get_hg_command(action, *args),
            user=self.resource.user,
            cwd=self.resource.name,
        )

    def should_update(self, *args):
        try:
            self.info("should-update", *args)
        except error.SystemError as e:
            self.logger.debug("'should-update' test has failed")
            self.logger.debug(e.stdout)
            return True
        return False

    def should_pull(self, *args):
        try:
            self.info("should-pull", *args)
        except error.SystemError as e:
            self.logger.debug("'should-pull' test has failed")
            self.logger.debug(e.stdout)
            return True
        return False

    def action(self, action, *args):
        self.change(
            ShellCommand(
                self.get_hg_command(action, *args),
                user=self.resource.user,
                cwd=self.resource.name,
            )
        )

    def apply(self):
        if not platform.exists("/usr/bin/hg"):
            self.raise_or_log(
                error.MissingDependency(
                    "'/usr/bin/hg' is not available; update your configuration to install mercurial?"
                )
            )
            return

        created = False
        changed = False

        self.change(
            EnsureDirectory(
                self.resource.name, self.resource.user, self.resource.group, 0o755
            )
        )

        if not platform.exists(os.path.join(self.resource.name, ".hg")):
            try:
                self.action("init")
            except error.SystemError:
                raise error.CheckoutError("Cannot initialise local repository.")
            created = True

        url = _inject_credentials(
            self.resource.repository,
            self.resource.scm_username,
            self.resource.scm_password,
        )

        try:
            self.change(
                EnsureFile(
                    os.path.join(self.resource.name, ".hg", "hgrc"),
                    contents=force_bytes(
                        hgrc % {"repository": url, "path": self.resource.name}
                    ),
                    user=self.resource.user,
                    group=self.resource.group,
                    mode=0o600,
                    sensitive=bool(self.resource.scm_password),
                )
            )
            # changed = changed or f.changed
        except error.SystemError:
            raise error.CheckoutError("Could not set the remote repository.")

        try:
            self.change(
                EnsureFile(
                    os.path.join(self.resource.name, ".hg", "should.py"),
                    contents=mercurial_ext,
                    user=self.resource.user,
                    group=self.resource.group,
                    mode=0o600,
                )
            )
            # changed = changed or f.changed
        except error.SystemError:
            raise error.CheckoutError("Could not setup mercurial idempotence extension")

        should_args = []
        if self.resource.branch:
            should_args.extend(["-b", self.resource.branch])
        if self.resource.tag:
            should_args.extend(["-t", self.resource.tag])

        if created or self.should_pull(*should_args):
            try:
                self.action("pull", "--force")
                changed = True
            except error.SystemError:
                raise error.CheckoutError(
                    "Could not fetch changes from remote repository."
                )

        if created or self.should_update(*should_args):
            if self.resource.tag:
                args = [self.resource.tag]
            elif self.resource.branch:
                args = [self.resource.branch]
            else:
                args = []

            try:
                self.action("update", *args)
                changed = True
            except error.SystemError:
                raise error.CheckoutError("Could not update working copy.")

        return created or changed



# Copyright 2011-2014 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from fuselage.argument import Boolean, String
from fuselage.policy import Absent, Policy, Present
from fuselage.resource import Resource


class Package(Resource):

    """Represents an operating system package, installed and managed via the
    OS package management system. For example, to ensure these three packages
    are installed::

        Package:
            - name: apache2
            - name: zip
            - name: xsltproc

    """

    name = String()
    """ The name of the package. This can be a single package or a list can be
    supplied. """

    version = String()
    """ The version of the package, if only a single package is specified and
    the appropriate provider supports it (the Apt provider does not support
    it). """

    purge = Boolean(default=False)
    """ When removing a package, whether to purge it or not. """

    backend = String()
    """ On a normal system, this does not need to be set. Set to 'apt' to force
    the use of apt-get, and 'yum' to force the use of yum. """


class PackageInstallPolicy(Policy):

    """Install the specified package. If the package is already installed it
    will not be upgraded or changed. Your package upgrade and patching
    strategy should be independent of Yaybu in general.
    """

    resource = Package
    name = "install"
    default = True
    signature = (
        Present("name"),
        Absent("purge"),
    )


class PackageUninstallPolicy(Policy):

    """Uninstall the specified package, if it is installed."""

    resource = Package
    name = "uninstall"
    default = False
    signature = (
        Present("name"),
        Absent("version"),
    )



# Copyright 2011-2014 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from fuselage import error, platform, provider, resources
from fuselage.changes import ShellCommand


class Execute(provider.Provider):

    policies = (resources.execute.ExecutePolicy,)

    def check_unless(self):
        try:
            platform.check_call(
                command=self.resource.unless,
                user=self.resource.user,
                cwd=self.resource.cwd,
            )

        except error.InvalidUser as exc:
            self.raise_or_log(exc)

        except error.InvalidGroup as exc:
            self.raise_or_log(exc)

        except error.SystemError:
            return True

        return False

    def apply(self):
        creates = self.resource.creates
        if creates and platform.exists(creates):
            self.logger.debug(f"{self.resource.creates!r} exists, not executing")
            return False

        touch = self.resource.touch
        if touch and platform.exists(touch):
            self.logger.debug(f"{self.resource.touch!r} exists, not executing")
            return False

        if self.resource.unless and not self.check_unless():
            self.logger.debug(f"{self.resource.unless!r} passes, not executing")
            return False

        if self.resource.command:
            commands = [self.resource.command]
        else:
            commands = self.resource.commands

        for command in commands:
            self.change(
                ShellCommand(
                    command=command,
                    cwd=self.resource.cwd or None,
                    env=self.resource.env or None,
                    user=self.resource.user or None,
                    group=self.resource.group or None,
                    umask=self.resource.umask,
                    expected=self.resource.returncode,
                )
            )

        if self.resource.touch:
            self.change(ShellCommand(["touch", self.resource.touch]))

        return True



# Copyright 2011-2014 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from fuselage import error, platform, provider, resources
from fuselage.changes import EnsureFile, ShellCommand
from fuselage.utils import force_bytes


class File(provider.Provider):

    """Provides file creation using templates or static files."""

    policies = (resources.file.FileApplyPolicy,)

    def check_path(self, directory):
        if platform.isdir(directory):
            self.logger.debug("Prereq: %r is a directory" % directory)
            return
        frags = directory.split("/")
        path = "/"
        for i in frags:
            path = os.path.join(path, i)
            if not platform.exists(path):
                self.raise_or_log(
                    error.PathComponentMissing("Directory '%s' is missing" % path)
                )
            elif not platform.isdir(path):
                raise error.PathComponentNotDirectory(
                    "Path '%s' is not a directory" % path
                )

    def apply(self):
        name = self.resource.name

        self.check_path(os.path.dirname(name))

        if self.resource.source:
            if self.resource.source.startswith("bundle://"):
                import pkgutil

                loader = pkgutil.get_loader("fuselage")
                contents = loader.get_data("assets/" + self.resource.source[9:])
            else:
                with open(self.resource.source, "rb") as fp:
                    contents = fp.read()
        elif self.resource.contents is not None:
            contents = force_bytes(self.resource.contents)
        else:
            contents = None

        fc = EnsureFile(
            name,
            contents,
            self.resource.owner,
            self.resource.group,
            self.resource.mode,
            sensitive=self.resource.sensitive,
        )
        self.change(fc)

        return fc.changed


class RemoveFile(provider.Provider):
    policies = (resources.file.FileRemovePolicy,)

    def get_delete_command(self):
        if platform.platform == "win32":
            return ["cmd.exe", "/C", "DEL", "/Q", "/F"]
        return ["rm"]

    def apply(self):
        name = self.resource.name
        if platform.exists(name):
            if not platform.isfile(name):
                raise error.InvalidProvider("%s exists and is not a file" % name)
            self.change(ShellCommand(self.get_delete_command() + [self.resource.name]))
            changed = True
        else:
            self.logger.debug("File %s missing already so not removed" % name)
            changed = False
        return changed


class WatchFile(provider.Provider):
    policies = (resources.file.FileWatchedPolicy,)

    def apply(self):
        """Watched files don't have any policy applied to them"""
        return self.resource.hash() != self.resource._original_hash



# Copyright 2014 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import json
import logging
import logging.handlers
import sys


class LoggerAdapter(logging.LoggerAdapter):

    """
    The upstream LoggerAdapter stops log.info("foo", extra={}) from working.
    This subclass doesnt.
    """

    def log(self, level, msg, *args, **kwargs):
        msg, kwargs = self.process(msg, kwargs)
        self.logger.log(level, msg, *args, **kwargs)

    def isEnabledFor(self, level):
        # Python 3.4 breaks on adapters of adapters, which is lame.
        return self.logger.isEnabledFor(level)

    def process(self, msg, kwargs):
        if not "extra" in kwargs:
            kwargs["extra"] = {}
        kwargs["extra"].update(self.extra)
        return msg, kwargs


class ResourceFormatter(logging.Formatter):

    """Automatically add a header and footer to log messages about particular
    resources"""

    def format(self, record):
        formatted = logging.Formatter.format(self, record)
        if hasattr(record, "fuselage.diff"):
            formatted = "\n".join((formatted, getattr(record, "fuselage.diff")))
        return formatted


class ConsoleHandler(logging.StreamHandler):
    def __init__(self, stream=sys.stdout, level=logging.INFO):
        super().__init__(stream)
        self._level = level
        self._resource = None
        self._needs_footer = False
        self._needs_header = False

    def format(self, record):
        prefix = "| " if self._resource else ""
        formatted = super().format(record)
        return "\n".join(prefix + line for line in formatted.splitlines())

    def _render_resource_header(self):
        if isinstance(self._resource, str):
            header = self._resource
        else:
            header = self._resource.encode("utf-8")

        rl = len(header)
        if rl < 80:
            total_minuses = 77 - rl
            minuses = int(total_minuses / 2)
            leftover = total_minuses % 2
        else:
            minuses = 4
            leftover = 0

        self.stream.write(
            "/{} {} {}\n".format("-" * minuses, header, "-" * (minuses + leftover))
        )

        self._needs_header = False
        self._needs_footer = True

    def _render_resource_footer(self):
        self.stream.write("\\{}\n\n".format("-" * 79))
        self._needs_footer = False

    def handle(self, record):
        record_type = getattr(record, "fuselage.type", None)
        next_resource = getattr(record, "fuselage.resource", None)

        if record_type == "resource-start":
            self._resource = next_resource
            self._needs_header = True

        if record.levelno >= self._level:
            if self._resource and self._needs_header:
                self._render_resource_header()
            super().handle(record)

        if record_type == "resource-finish":
            if self._needs_footer:
                self._render_resource_footer()
            self._resource = None


class JSONHandler(logging.StreamHandler):

    """
    Output all log info (for specified verbosity) to stdout in JSON. This
    allows parent process to track and report on progress.
    """

    def format(self, record):
        # FIXME: Is implementing a "formatter" enough??
        return json.dumps(record)


class SysLogHandler(logging.handlers.SysLogHandler):

    """
    We output an audit log to syslog. This only contains log entries that are
    marked as changes.
    """

    def filter(self, record):
        if "fuselage.change" not in record.__dict__:
            return 0
        return 1


def configure(verbosity=logging.INFO, json=False, force=False):
    root = logging.getLogger()
    root.setLevel(logging.DEBUG)

    if len(root.handlers) != 0:
        if not force:
            return
        root.handlers[:] = []

    if json:
        root.addHandler(JSONHandler(sys.stdout))
    else:
        handler = ConsoleHandler(sys.stdout, level=verbosity)
        handler.setFormatter(ResourceFormatter())
        root.addHandler(handler)

    # root.addHandler(SysLogHandler())



# Copyright 2011-2014 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import logging

from fuselage import error, log
from fuselage.resource import Resource, ResourceType

logger = logging.getLogger(__name__)


class ResourceBundle:

    """An ordered, indexed collection of resources. Pass in a specification
    that consists of scalars, lists and dictionaries and this class will
    instantiate the appropriate resources into the structure."""

    BUNDLE_VERSION = 1

    def __init__(self):
        self.clear()

    @classmethod
    def from_iterator(cls, iterator):
        bun = cls()
        if iterator:
            bun.extend(iterator)
        return bun

    def __len__(self):
        return len(self.resources)

    def __getitem__(self, key):
        return self._index_by_id[key]

    def clear(self):
        self.resources = []
        self._index_by_id = {}

    def get_resource_by_name(self, target):
        for res in self.resources:
            if res.id == target:
                return res
        else:
            raise KeyError("No such resource by name '%s'" % target)

    def dump(self, builder, fp):
        obj = self._serialize_bundle(builder)
        json.dump(obj, fp)

    def dumps(self, builder):
        obj = self._serialize_bundle(builder)
        return json.dumps(obj)

    def _serialize_bundle(self, builder):
        obj = {"version": self.BUNDLE_VERSION}
        resources = obj["resources"] = []
        for r in self.resources:
            if not getattr(r, "_implicit", False):
                resources.append(r.serialize(builder))
        return obj

    def load(self, fp):
        obj = json.load(fp)
        return self._load_bundle(obj)

    def loads(self, s):
        obj = json.loads(s)
        return self._load_bundle(obj)

    def _load_bundle(self, obj):
        if not isinstance(obj, dict):
            raise error.ParseError("Bundle is not a dictionary")

        if "version" not in obj:
            raise error.ParseError("Bundle version is invalid")

        if obj["version"] > self.BUNDLE_VERSION:
            raise error.ParseError("Bundle version is too new")

        if "resources" not in obj:
            raise error.ParseError("Bundle doesn't have a resources key")

        if not isinstance(obj["resources"], list):
            raise error.ParseError("Bundle's resource list is not a list")

        for resource in obj["resources"]:
            if not isinstance(resource, dict):
                raise error.ParseError("Not a valid resource definition")

            if len(resource) != 1:
                raise error.ParseError(
                    "Wrong number of keys in outer resource definition"
                )

            typename = list(resource.keys())[0]
            instances = resource[typename]
            if isinstance(instances, dict):
                instances = [instances]

            for instance in instances:
                self.create(typename, **instance)

    def create(self, typename, **kwargs):
        try:
            kls = ResourceType.resources[typename]
        except KeyError:
            raise error.ParseError("There is no resource type of '%s'" % typename)

        resource = kls(**kwargs)

        return self.add(resource)

    def add(self, resource: Resource):
        if resource.typed_id in self._index_by_id:
            raise error.ParseError("Resources cannot be defined multiple times")

        # Create implicit File[] nodes for any watched files
        for watched in resource.changes:
            w = self.create("File", **{"name": watched, "policy": "watched"})
            w._original_hash = None
            w._implicit = True

        resource.bind(self)
        self.resources.append(resource)
        self._index_by_id[resource.typed_id] = resource
        return resource

    def extend(self, iterator):
        while True:
            try:
                resource = next(iterator)
            except StopIteration:
                break

            try:
                self.add(resource)
            except Exception as e:
                throw = getattr(iterator, "throw", None)
                if callable(throw):
                    throw(e)
                else:
                    raise

    def apply(self, runner):
        """Apply the resources to the system, using the provided context and
        overall configuration."""
        for resource in self.resources:
            if hasattr(resource, "_original_hash"):
                resource._original_hash = resource.hash()

        something_changed = False
        mylen = len(self.resources)
        for i, resource in enumerate(self.resources, start=1):
            resource_log = log.LoggerAdapter(
                logger, {"fuselage.resource": resource.typed_id}
            )

            resource_log.debug(
                "Started applying '%r' (%d of %d)" % (resource, i, mylen),
                extra={"fuselage.type": "resource-start"},
            )
            try:
                if resource.apply(runner):
                    resource_log.debug(f"'{resource!r}' made changes")
                    something_changed = True
            finally:
                resource_log.debug(
                    f"Finished applying '{resource!r}'",
                    extra={"fuselage.type": "resource-finish"},
                )

        if not something_changed:
            raise error.NothingChanged()



import os
import time
from typing import Optional

import paramiko

from .builder import build
from .bundle import ResourceBundle


def iter_chunks(channel: paramiko.Channel):
    while not channel.exit_status_ready():
        while channel.recv_ready():
            yield channel.recv(1024)
        time.sleep(0.1)

    while channel.recv_ready():
        yield channel.recv(1024)


def iter_lines(channel: paramiko.Channel):
    buffer = bytearray()
    for chunk in iter_chunks(channel):
        buffer.extend(chunk)

        while b"\n" in buffer:
            line, buffer = buffer.split(b"\n", 1)
            yield line.decode("utf-8")


def execute_via_ssh(
    transport: paramiko.Transport,
    bundle: ResourceBundle,
    dry_run: bool = False,
    username: Optional[str] = "root",
    sudo_password: Optional[str] = None,
):
    """
    transport = paramiko.Transport(("localhost", 22))
    transport.connect(
        username="ubuntu",
        password="password55",
    )
    execute(transport, bundle, "root", "mysudopassword")
    """
    payload = build(bundle)

    sftp = transport.open_sftp_client()
    sftp.chdir(".")

    path = os.path.join(sftp.getcwd(), ".payload.pex")

    command_parts = [path, "--resume"]

    if dry_run:
        command_parts.append("--simulate")

    command = " ".join(command_parts)

    channel = transport.open_session()
    channel.get_pty()
    channel.set_combine_stderr(1)

    try:
        sftp.putfo(payload, path)
        sftp.chmod(path, 0o755)

        try:
            if username and username != transport.get_username():
                command = f"sudo -u {username} {command}"
                channel.exec_command(command)

                if sudo_password:
                    first_line = channel.recv(1024)

                    if b"[sudo]" not in first_line and not first_line.startswith(
                        b"Password:"
                    ):
                        raise RuntimeError(f"Expected sudo, got {first_line!r}")

                    channel.sendall((sudo_password + "\n").encode("utf-8"))
            else:
                channel.exec_command(path)

            channel.shutdown_write()

            try:
                for line in iter_lines(channel):
                    print(line)

                return channel.recv_exit_status()
            finally:
                channel.close()
        finally:
            sftp.remove(path)
    finally:
        sftp.close()



# Copyright 2011-2014 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Resources for handling the creation and removal of files. These deal with
both the metadata associated with the file (for example owner and permission)
and the contents of the files themselves. """

from fuselage.argument import Boolean, File, FullPath, Octal, String
from fuselage.defaults import get_default_group, get_default_user
from fuselage.policy import NAND, Absent, Policy, Present
from fuselage.resource import Resource


class File(Resource):

    """A provider for this resource will create or amend an existing file to
    the provided specification.

    For example, the following will create the /etc/hosts file based on a static local file::

        File:
          name: /etc/hosts
          owner: root
          group: root
          mode: 644
          static: my_hosts_file

    The following will create a file using a jinja2 template, and will back up
    the old version of the file if necessary::

        File:
          name: /etc/email_addresses
          owner: root
          group: root
          mode: 644
          template: email_addresses.j2
          template_args:
              foo: foo@example.com
              bar: bar@example.com
          backup: /etc/email_addresses.{year}-{month}-{day}

    """

    name = FullPath()
    """The full path to the file this resource represents."""

    owner = String(default=lambda r: get_default_user())
    """A unix username or UID who will own created objects. An owner that
    begins with a digit will be interpreted as a UID, otherwise it will be
    looked up using the python 'pwd' module."""

    group = String(default=lambda r: get_default_group(r.owner))
    """A unix group or GID who will own created objects. A group that begins
    with a digit will be interpreted as a GID, otherwise it will be looked up
    using the python 'grp' module."""

    mode = Octal(default=0o644)
    """A mode representation as an octal. This can begin with leading zeros if
    you like, but this is not required. DO NOT use yaml Octal representation
    (0o666), this will NOT work."""

    contents = String(default=None)
    """ A complete string to write into a file """

    source = File()
    """A file that will be applied to this resource. """

    sensitive = Boolean(default=False)

    def hash(self):
        from fuselage import platform

        try:
            return platform.stat(self.name).st_mtime
        except OSError:
            return 0


class FileApplyPolicy(Policy):

    """Create a file and populate it's contents if required.

    You must provide a name.

    You may provide one of template, static, or encrypted to act as a file source.
    """

    resource = File
    name = "apply"
    default = True
    signature = (Present("name"), NAND(Present("contents"), Present("source")))


class FileRemovePolicy(Policy):

    """Delete a file if it exists. You should only provide the name in this
    case."""

    resource = File
    name = "remove"
    default = False
    signature = (
        Present("name"),
        Absent("owner"),
        Absent("group"),
        Absent("mode"),
    )


class FileWatchedPolicy(Policy):

    """Watches a file to see if it changes when a resource a file.

    This policy is used internally and shouldn't be used directly.
    """

    resource = File
    name = "watched"
    default = False
    signature = FileRemovePolicy.signature



# Copyright 2011-2014 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from fuselage import error, platform, provider, resources
from fuselage.changes import ShellCommand


def is_installed(resource):
    # work out if the package is already installed
    command = ["dpkg-query", "-W", "-f='${Status}'", resource.name]

    try:
        stdout, stderr = platform.check_call(command)
    except error.SystemError as exc:
        if exc.returncode == 1:
            return False
        # if the return code is anything but zero or one, we have a problem
        raise error.PackageError(
            f"{resource} search failed with return code {exc.returncode}"
        )

    # if the return code is 0, dpkg is aware of the package
    if "install ok installed" in stdout:
        return True

    return False


class AptInstall(provider.Provider):

    policies = (resources.package.PackageInstallPolicy,)

    @classmethod
    def isvalid(self, policy, resource):
        if resource.backend and resource.backend != "apt":
            return False
        return True

    def apply(self):
        if is_installed(self.resource):
            self.logger.debug("Package already installed; nothing to do.")
            return False

        env = {
            "DEBIAN_FRONTEND": "noninteractive",
        }

        # the search returned 1, package is not installed, continue and install
        # it
        command = [
            "apt-get",
            "install",
            "-q",
            "-y",
            "--force-yes",
            "-o",
            "DPkg::Options::=--force-confdef",
            self.resource.name,
        ]

        try:
            self.change(ShellCommand(command, env=env))
        except error.SystemError as exc:
            if exc.returncode == 100:
                try:
                    self.change(
                        ShellCommand(["apt-get", "update", "-q", "-y"], env=env)
                    )
                    self.change(ShellCommand(command, env=env))
                except error.SystemError as exc:
                    raise error.PackageError(
                        "%s with what looked like a recoverable error, but it wasn't (return code %d)"
                        % (self.resource, exc.returncode)
                    )
            else:
                raise error.PackageError(
                    "%s failed with return code %d" % (self.resource, exc.returncode)
                )

        return True


class AptUninstall(provider.Provider):

    policies = (resources.package.PackageUninstallPolicy,)

    @classmethod
    def isvalid(self, policy, resource):
        if resource.backend and resource.backend != "apt":
            return False
        return True

    def apply(self):
        if not is_installed(self.resource):
            self.logger.debug(
                "Package '%s' is already uninstalled, not removing."
                % self.resource.name
            )
            return False

        env = {
            "DEBIAN_FRONTEND": "noninteractive",
        }

        command = ["apt-get", "remove", "-q", "-y"]
        if self.resource.purge:
            command.append("--purge")
        command.append(self.resource.name)

        try:
            self.change(ShellCommand(command, env=env))
        except error.SystemError as exc:
            raise error.PackageError(
                "%s failed to uninstall with return code %d"
                % (self.resource, exc.returncode)
            )

        return True



# Copyright 2011-2014 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from fuselage import error, log, policy
from fuselage.argument import (
    Argument,
    List,
    PolicyArgument,
    String,
    SubscriptionArgument,
)
from fuselage.utils import force_str

logger = logging.getLogger(__name__)


class ResourceType(type):

    """Keeps a registry of resources as they are created, and provides some
    simple access to their arguments."""

    resources = {}

    def __new__(meta, class_name, bases, new_attrs):
        new_attrs.setdefault("__resource_name__", class_name)
        cls = type.__new__(meta, class_name, bases, new_attrs)

        cls.__args__ = {}

        for b in bases:
            if hasattr(b, "__args__"):
                cls.__args__.update(b.__args__)

        cls.policies = AvailableResourcePolicies()

        if class_name != "Resource":
            if cls.__resource_name__ in meta.resources:
                raise error.ParseError(
                    "Redefinition of resource %s" % cls.__resource_name__
                )
            else:
                meta.resources[cls.__resource_name__] = cls

        for key, value in new_attrs.items():
            if isinstance(value, Argument):
                cls.__args__[key] = value

        return cls

    @classmethod
    def clear(self):
        self.resources = {}


class AvailableResourcePolicies(dict):

    """A collection of the policies available for a resource, with some logic
    to work out which of them is the one and only default policy."""

    def default(self):
        default = [p for p in self.values() if p.default]
        if default:
            return default[0]
        else:
            return policy.NullPolicy


class Resource(metaclass=ResourceType):

    """A resource represents a resource that can be configured on the system.
    This might be as simple as a symlink or as complex as a database schema
    migration. Resources have policies that represent how the resource is to
    be treated. Providers are the implementation of the resource policy.

    Resource definitions specify the complete set of attributes that can be
    configured for a resource. Policies define which attributes must be
    configured for the policy to be used.

    """

    policies = AvailableResourcePolicies()
    """ A dictionary of policy names mapped to policy classes (not objects).

    These are the policies for this resource class.

    Here be metaprogramming magic.

    Dynamically allocated as Yaybu starts up this is effectively static once
    we're up and running. The combination of this attribute and the policy
    argument below is sufficient to determine which provider might be
    appropriate for this resource.

    """

    policy = PolicyArgument()
    """ The list of policies provided by configuration. This is an argument
    like any other, but has a complex representation that holds the conditions
    and options for the policies as specified in the input file. """

    id = String(default=lambda r: r.implicit_id)

    watches = SubscriptionArgument()

    changes = List(default=[])
    """ A list of files to monitor while this resource is applied

    The file will be hashed before and after a resource is applied.
    If the hash changes, then it will be like a policy has been applied
    on that file.

    For example::

        resources.append:
          - Execute:
              name: buildout-foobar
              command: buildout2.6
              watch:
                - /var/local/sites/foobar/apache/apache.cfg

          - Service:
              name: apache2
              policy:
                restart:
                  when: watched
                  on: File[/var/local/sites/foobar/apache/apache.cfg]
    """

    def __init__(self, **kwargs):
        """Takes a reference to a Yay AST node"""
        self.observers = list()

        for key, value in kwargs.items():
            if key not in self.__args__:
                raise error.ParseError(
                    f"'{key}' is not a valid option for this resource"
                )
            setattr(self, key, value)

        self.policy.validate()
        self.policy.get_provider()

        if not self.id:
            raise error.ParseError(
                ("{} is not explicitly named and name cannot be implied").format(
                    self.__resource_name__
                )
            )

    @classmethod
    def get_argument_names(klass):
        yield from klass.__args__

    def serialize(self, builder=None):
        """Return all argument names and values in a dictionary. If an
        argument has no default and has not been set, it's value in the
        dictionary will be None."""
        retval = {}
        for name, arg in self.__args__.items():
            if arg.present(self):
                retval[name] = arg.serialize(self, builder=builder)
        return {self.__resource_name__: retval}

    def register_observer(self, when, resource, policy):
        logger.debug(f"{self!r} is being observed by {resource!r} for {when}")
        self.observers.append(resource)

    def apply(self, runner):
        """Apply the provider for the selected policy, and then fire any
        events that are being observed."""

        adapter = log.LoggerAdapter(logger, {"fuselage.resource": self.id})

        if self.watches and not runner.state.is_trigger_set(self):
            adapter.debug(
                "Skipping resource apply as subscribed to triggers that are not set"
            )
            return False

        provider = self.policy.get_provider()(self, runner)
        changed = provider.apply()
        runner.state.unset_trigger(self)
        if changed:
            self.fire_event(runner)
        return changed

    def fire_event(self, context):
        """Apply the appropriate policies on the resources that are observing
        this resource for the firing of a policy."""
        logger.debug(f"Sending triggers from {self}")
        for resource in self.observers:
            logger.debug(f"Sending trigger from {self!r} to {resource!r}")
            context.state.set_trigger(resource)

    def bind(self, resources):
        """Bind this resource to all the resources on which it triggers.
        Returns a list of the resources to which we are bound."""
        bound = []
        if self.watches:
            for trigger in self.watches:
                bound.append(trigger.bind(resources, self))
        return bound

    @property
    def implicit_id(self):
        if getattr(self, "name", None):
            return force_str(self.name)
        return None

    @property
    def typed_id(self):
        name = self.id
        if not name:
            raise error.ParseError("Resource is not named")
        classname = getattr(self, "__resource_name__", self.__class__.__name__)
        return f"{classname}[{name}]"

    def __repr__(self):
        return self.typed_id



# Copyright 2011-2014 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from fuselage.argument import FullPath, Octal, String
from fuselage.defaults import get_default_group, get_default_user
from fuselage.policy import Policy, Present
from fuselage.resource import Resource


class Checkout(Resource):

    """This represents a "working copy" from a Source Code Management system.
    This could be provided by, for example, Subversion or Git remote
    repositories.

    Note that this is '*a* checkout', not 'to checkout'. This represents the
    resource itself on disk. If you change the details of the working copy
    (for example changing the branch) the provider will execute appropriate
    commands (such as `svn switch`) to take the resource to the desired state.
    """

    name = FullPath()
    """ The full path to the working copy on disk. """

    repository = String()
    """ The identifier for the repository - this could be an http url for
    subversion or a git url for git, for example. """

    branch = String()
    """ The name of a branch to check out, if required. """

    tag = String()
    """ The name of a tag to check out, if required. """

    revision = String()
    """ The revision to check out or move to. """

    scm = String()
    """ The source control management system to use, e.g. subversion, git. """

    scm_username = String()
    """ The username for the remote repository """

    scm_password = String()
    """ The password for the remote repository. """

    user = String(default=lambda r: get_default_user())
    """ The user to perform actions as, and who will own the resulting files. """

    group = String(default=lambda r: get_default_group(r.user))
    """ The group to perform actions as. """

    mode = Octal(default="755")
    """A mode representation as an octal. This can begin with leading zeros if
    you like, but this is not required. DO NOT use yaml Octal representation
    (0o666), this will NOT work."""


class CheckoutSyncPolicy(Policy):

    """Synchronise the working copy with the remote SCM location. This is
    done by performing appropriate "switch" and "update" operations such that
    the working copy ends up in the correct state. This will never commit
    changes from the working copy - it is intended to track a remote location
    only."""

    resource = Checkout
    name = "sync"
    default = True
    signature = (
        Present("name"),
        Present("repository"),
        Present("scm"),
    )


class CheckoutExportPolicy(Policy):

    """Perform an export of the remote location, as for example "svn export".
    This means this is not a working copy of the remote location, it is
    instead a clean copy of the specified source material."""

    resource = Checkout
    name = "export"
    default = False
    signature = (
        Present("name"),
        Present("repository"),
        Present("scm"),
    )



# Copyright 2014 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import hashlib
import io
import modulefinder
import os
import pkgutil
import zipfile

from fuselage.bundle import ResourceBundle

MAIN_PY = """
import logging
import sys
from fuselage import resources, error, log
from fuselage.runner import BundledRunner
if __name__=="__main__":
    log.configure()
    logger = logging.getLogger("fuselage")
    try:
        r = BundledRunner.setup_from_cmdline()
        r.run()
    except error.Error as e:
        logger.error(e)
        sys.exit(e.returncode)
"""


class Builder:
    def __init__(self, zfp):
        self.zipfile = zfp

    @classmethod
    def write_to(cls, fp):
        fp.write(b"#!/usr/bin/env python3\n")
        obj = cls(zipfile.ZipFile(fp, "w", compression=zipfile.ZIP_DEFLATED))
        obj.fp = fp
        return obj

    @classmethod
    def write_to_path(cls, path):
        return cls.write_to(open(path, "wb"))

    def add_resource_blob(self, payload):
        name = hashlib.sha1(payload).hexdigest()
        self.zipfile.writestr(os.path.join("assets", name), payload)
        return "bundle://" + name

    def embed_resource_bundle(self, bundle: ResourceBundle):
        data = bundle.dumps(self)
        self.zipfile.writestr("resources.json", data)

    def embed_fuselage_runtime(self):
        finder = modulefinder.ModuleFinder()

        co = compile(MAIN_PY, "__main__.py", "exec")
        m = finder.add_module("__main__")
        m.__file__ = "__main__.py"
        finder.scan_code(co, m)

        for name, mod in finder.modules.items():
            if not name.startswith("fuselage"):
                continue

            # Use pkgutil to get the code - this is zipsafe so will work even if
            # running from a py2exe type binary installation.
            basename = os.path.basename(mod.__file__)
            path_parts = out_parts = list(name.split("."))
            if basename == "__init__.py":
                path_parts.append("__init__.py")
            elif not "." in name:
                path_parts = [name, name + ".py"]
                out_parts = [name + ".py"]
            else:
                path_parts[-1] += ".py"

            code = pkgutil.get_data(path_parts[0], os.sep.join(path_parts[1:]))
            self.zipfile.writestr(os.path.join(*out_parts), code)

        self.zipfile.writestr("__init__.py", "")
        self.zipfile.writestr("__main__.py", MAIN_PY)

    def close(self):
        self.zipfile.close()


def build(bundle: ResourceBundle, name: str = "payload.pex") -> io.BytesIO:
    buffer = io.BytesIO()
    buffer.name = name

    bu = Builder.write_to(buffer)
    bu.embed_fuselage_runtime()
    bu.embed_resource_bundle(bundle)
    bu.close()

    buffer.seek(0)
    return buffer



# Copyright 2014 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from fuselage import platform


def get_default_user():
    if not platform.getpwuid:
        return ""
    return platform.getpwuid(platform.getuid()).pw_name


def get_default_group(user):
    if not platform.getpwnam:
        return ""
    try:
        gid = platform.getpwnam(user).pw_gid
        return platform.getgrgid(gid).gr_name
    except KeyError:
        return None



# Copyright 2011-2014 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import difflib
import string

from fuselage import platform
from fuselage.changes import base
from fuselage.utils import force_str

from .attributes import AttributeChanger
from .execute import ShellCommand

# FIXME: Set mode of file before writing to it

PRINTABLE = string.printable.encode("utf-8")


def binary_buffers(*buffers):
    """Check all of the passed buffers to see if any of them are binary. If
    any of them are binary this will return True."""

    def is_printable(value):
        return len(buff) == sum(1 for c in buff if c in PRINTABLE)

    for buff in buffers:
        if not buff:
            continue
        if not is_printable(buff):
            return True
    return False


class EnsureContents(base.Change):

    """Apply a content change to a file in a managed way. Simulation mode is
    catered for. Additionally the minimum changes required to the contents are
    applied, and logs of the changes made are recorded."""

    def __init__(self, filename, contents, sensitive=False):
        self.filename = filename
        self.current = b""
        self.contents = contents
        self.changed = False
        self.sensitive = sensitive

    def diff(self, context, note, previous, replacement):
        extra = {}
        if self.sensitive:
            extra["fuselage.diff"] = "No diff; sensitive file contents"
        elif not binary_buffers(previous, replacement):
            previous = force_str(previous)
            replacement = force_str(replacement)
            diff = "".join(
                difflib.unified_diff(previous.splitlines(1), replacement.splitlines(1))
            )
            extra["fuselage.diff"] = diff
        else:
            extra["fuselage.diff"] = "No diff; binary files"
        context.changelog.critical(note, extra=extra)

    def empty_file(self, context):
        """Write an empty file if none exists"""
        if not platform.exists(self.filename):
            context.changelog.debug("Creating empty file %r" % self.filename)
            context.change(ShellCommand(["touch", self.filename]))
            self.changed = True

    def overwrite_existing_file(self, context):
        """Change the content of an existing file"""
        self.current = platform.get(self.filename)
        if self.current != self.contents:
            self.diff(context, "Changing existing file", self.current, self.contents)
            if not context.simulate:
                platform.put(self.filename, self.contents)
            self.changed = True
        else:
            context.logger.debug("Not changing content of file %r" % self.filename)

    def write_new_file(self, context):
        """Write contents to a new file."""
        self.diff(context, "Writing new file", b"", self.contents)
        if not context.simulate:
            platform.put(self.filename, self.contents)
        self.changed = True

    def write_file(self, context):
        """Write to either an existing or new file"""
        if platform.exists(self.filename):
            self.overwrite_existing_file(context)
        else:
            self.write_new_file(context)

    def apply(self, context):
        if self.contents is None:
            self.empty_file(context)
        else:
            self.write_file(context)
        return self.changed


class EnsureFile(base.Change):
    def __init__(self, filename, contents, user, group, mode, sensitive=False):
        self.filename = filename
        self.contents = contents
        self.user = user
        self.group = group
        self.mode = mode
        self.changed = False
        self.sensitive = sensitive

    def apply(self, context):
        """Apply the changes necessary to the file contents."""
        fc = EnsureContents(self.filename, self.contents, sensitive=self.sensitive)
        context.change(fc)

        ac = AttributeChanger(self.filename, self.user, self.group, self.mode)
        context.change(ac)

        self.changed = fc.changed or ac.changed
        return self



# Copyright 2011-2014 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os

from fuselage import error, platform
from fuselage.utils import force_str


class EventState:

    """
    Represents the current state of events

    Createw a JSON state file on the target system. fuselage is idempotent - it
    doesn't rely on this to know if it's applied your changes. It relies on it
    on it to keep track of watch triggers.

    For example, if you make a change to a File, fuselage doesn't need this
    state data to make sure the file is updated. But if you have an Execute
    that watches the File then this state date will ensure the Execute is
    trigger when resuming from failure.
    """

    overrides = {}
    """ A mapping of resource ids to the overridden policy name for that
    resource, if there is one. """

    def __init__(self, save_file, simulate, load=False):
        self.save_file = save_file
        self.simulate = simulate
        self.loaded = not load
        self.overrides = {}
        self.simulate = simulate

        # FIXME
        self.resume = True
        self.no_resume = False

    def load(self):
        if self.loaded:
            return
        if platform.exists(self.save_file):
            self.overrides = json.loads(force_str(platform.get(self.save_file)))
        self.loaded = True

    def set_trigger(self, resource):
        self.load()
        self.overrides[resource.id] = "*"
        self.save()

    def unset_trigger(self, resource):
        self.load()
        if resource.id in self.overrides:
            del self.overrides[resource.id]
            self.save()

    def is_trigger_set(self, resource):
        self.load()
        return resource.id in self.overrides

    def open(self):
        if not self.simulate:
            save_parent = os.path.realpath(os.path.join(self.save_file, os.path.pardir))
            if not platform.exists(save_parent):
                platform.makedirs(save_parent)

        if not platform.exists(self.save_file):
            return

        if self.resume:
            self.loaded = False
            return
        elif self.no_resume:
            if not self.simulate:
                platform.unlink(self.save_file)
            self.loaded = True
            return

        raise error.SavedEventsAndNoInstruction()

    def success(self):
        if not self.simulate and platform.exists(self.save_file):
            platform.unlink(self.save_file)

    def save(self):
        if not self.simulate:
            platform.put(self.save_file, json.dumps(self.overrides))



# Copyright 2011-2014 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .checkout import Checkout
from .directory import Directory
from .execute import Execute
from .file import File
from .group import Group
from .line import Line
from .link import Link
from .mount import Mount
from .package import Package
from .patch import Patch
from .service import Service
from .special import Special
from .user import User

__all__ = [
    "Checkout",
    "Directory",
    "Execute",
    "File",
    "Group",
    "Line",
    "Link",
    "Mount",
    "Package",
    "Patch",
    "Service",
    "Special",
    "User",
]



"""

Classes that represent errors within fuselage.

What is listed here are the exceptions raised within Python, with an
explanation of their meaning. If you wish to detect a specific error on
invocation, you can do so via the return code of the fuselage process.

All fuselage errors have a returncode, which is returned from the fuselage program
if these errors occur. This is primarily for the test harness, but feel free
to rely on these, they should be stable.

A returncode of less than 128 is an error from within the range specified in
the errno library, which contains the standard C error codes.

These may have been actually returned from a shell command, or they may be
based on our interpretation of the failure mode they represent. Resources will
define the errors they may return. """

import errno


class Error(Exception):
    """Base class for all fuselage specific exceptions."""

    returncode = 253

    def __init__(self, msg=""):
        if msg:
            self.msg = msg

    def __str__(self):
        return f"{self.__class__.__name__}: {self.msg}"


class ParseError(Error):
    """Root of exceptions that are caused by an error in input."""

    returncode = 128


class BindingError(Error):
    """An error during policy binding."""

    returncode = 129


class ExecutionError(Error):
    """Root of exceptions that are caused by execution failing in an unexpected way."""

    returncode = 130


class PackageError(ExecutionError):
    """Error whilst performing a packaging error"""

    returncode = 131


class CommandError(ExecutionError):
    """A command from the Execute provider did not return the expected return
    code."""

    returncode = 133


class NoValidPolicy(ParseError):
    """There is no valid policy for the resource."""

    returncode = 135


class NonConformingPolicy(ParseError):
    """A policy has been specified, or has been chosen by default, but the
    parameters provided for the resource do not match those required for the
    policy. Check the documentation to ensure you have provided all required
    parameters."""

    returncode = 136


class NoSuitableProviders(ParseError):
    """There are no suitable providers available for the policy and resource
    chosen. This may be because a provider has not been written for this
    Operating System or service, or it may be that you have not specified the
    parameters correctly."""

    msg = "There is no uitable provider for this resource/policy"
    returncode = 137


class TooManyProviders(ParseError):
    """More than one provider matches the specified resource, and Yaybu is unable to choose between them."""

    returncode = 138


class InvalidProvider(ExecutionError):
    """A provider is not valid. This is detected before any changes have been
    applied."""

    returncode = 139


class InvalidGroup(ExecutionError):
    """The specified user group does not exist."""

    returncode = 140


class InvalidUser(ExecutionError):
    """The specified user does not exist."""

    returncode = 141


class OperationFailed(ExecutionError):
    """A general failure of an operation. For example, we tried to create a
    symlink, everything appeared to work but then a link does not exist. This
    should probably never happen."""

    returncode = 142


class BinaryMissing(ExecutionError):
    """A specific error for an expected binary (ln, rm, etc.) not being
    present where expected."""

    returncode = 143


class DanglingSymlink(ExecutionError):
    """The destination of a symbolic link does not exist."""

    returncode = 144


class UserAddError(ExecutionError):
    """An error from the useradd command. It has a bunch of error codes of
    it's own."""

    returncode = 145


class PathComponentMissing(ExecutionError):
    """A component of the path is not present"""

    returncode = 146


class PathComponentNotDirectory(ExecutionError):
    """A component of the path is in fact not a directory"""

    returncode = 147


class SavedEventsAndNoInstruction(Error):
    """There is a saved events file and the user has not decided what to do
    about it. Invoke with --resume or --no-resume."""

    returncode = 148
    msg = "There is a saved events file - you need to specify --resume or --no-resume"


class MissingAsset(ExecutionError):
    """An asset referenced by a resource could not be found on the Yaybu
    search path."""

    returncode = 149


class CheckoutError(Error):
    """An insurmountable problem was encountered during checkout"""

    returncode = 150


class Incompatible(Error):
    """An incompatibility was detected and execution can't continue"""

    returncode = 151


class MissingDependency(ExecutionError):
    """A dependency required for a feature or provider is missing"""

    returncode = 152


class UnmodifiedAsset(ExecutionError):
    """An asset was requested unnecesarily. This indicates an error in cache
    handling and should be filed as a bug against Yaybu."""

    returncode = 153


class NothingChanged(ExecutionError):
    """Not really an error, but we need to know if this happens for our
    tests. This exception is never really raised, but it's useful to keep the
    error code here!"""

    msg = "No changes have been applied"
    returncode = 254


class SystemError(ExecutionError):
    """An error represented by something in the errno list."""

    def __init__(self, returncode, stdout=None, stderr=None):
        # if the returncode is not in errno, this will blow up.
        try:
            self.msg = errno.errorcode[returncode]
        except KeyError:
            self.msg = "Exit code %s" % returncode
        self.returncode = returncode
        self.stdout = stdout
        self.stderr = stderr



# Copyright 2014 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import inspect
import os
import subprocess

import ConfigParser
import StringIO

from fuselage import builder, bundle, error

try:
    from fabric import tasks, utils
    from fabric.api import env, settings
    from fabric.operations import put, sudo
except SyntaxError:
    raise ImportError(
        "Fabric cannot be imported due to syntax errors. Are you using a supported version of python?"
    )


class Loader:
    def __init__(self, dirname):
        self._dirname = dirname

    @property
    def dirname(self):
        return self._dirname() if callable(self._dirname) else self._dirname

    def exists(self, path):
        return os.path.exists(os.path.join(self.dirname, path))

    def template(self, path, ctx):
        from jinja2 import Environment, FileSystemLoader

        e = Environment(
            loader=FileSystemLoader(self.dirname), line_statement_prefix="%"
        )
        return e.get_template(path).render(ctx) + "\n"

    def static(self, path):
        with open(os.path.join(self.dirname, path), "rb") as fp:
            return fp.read()

    def decrypt(self, path):
        data = self.static(path)
        environ = os.environ.copy()
        if "GPG_TTY" not in environ and os.path.exists("/proc/self/fd/0"):
            environ["GPG_TTY"] = os.readlink("/proc/self/fd/0")
        p = subprocess.Popen(
            ["gpg", "--use-agent", "--batch", "-d"],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            stdin=subprocess.PIPE,
            env=environ,
        )
        stdout, stderr = p.communicate(data)
        if p.returncode != 0:
            error_message = "Unable to decrypt data '%s'." % path
            if "GPG_AGENT_INFO" not in environ:
                error_message += (
                    " GPG Agent not running so your GPG key may not be available."
                )
            utils.error(error_message)
            raise RuntimeError(error_message)
        return stdout


class Environment(ConfigParser.ConfigParser):
    def __init__(self, environment, dirname=None):
        super().__init__()
        self.loader = Loader(
            dirname or os.path.join(os.path.dirname(env.real_fabfile), "environments")
        )
        self.read("default.cfg")
        self.read("default-secrets.cfg.gpg")
        self.read(environment + ".cfg")
        self.read(environment + "-secrets.cfg.gpg")

    def read(self, path, required=False):
        fn = self.loader.decrypt if path.endswith(".gpg") else self.loader.static
        if not self.loader.exists(path):
            if required:
                utils.error("Configuration file '%s' not found" % path)
                raise RuntimeError
            return
        self.readfp(StringIO.StringIO(fn(path)))

    def iter_sections_starting(self, starting):
        for section in self.sections():
            if section.startswith(starting):
                length = len(starting)
                yield section[length:], dict(self.items(section))


class FuselageMixin:

    arguments = []

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.arguments = list(self.arguments)
        for arg in inspect.getargspec(self.wrapped).args:
            if arg not in ("bundle",):
                self.arguments.append(arg)

        self.kwargs = {}
        for k, v in kwargs.items():
            if k in self.arguments:
                self.kwargs[k] = v

    def get_bundle_stream(self, *args, **kwargs):
        try:
            bun = bundle.ResourceBundle()
            bun.extend(self.wrapped(bun, *args, **kwargs))
        except error.Error as e:
            utils.error(str(e), exception=e)
            return

        return bun

    def apply_bundle(self, buffer, *args, **kwargs):
        raise NotImplementedError(self.apply_bundle)

    def run(self, *args, **kwargs):
        unsupported = set(kwargs.keys()).difference(set(self.arguments))
        if unsupported:
            utils.error(
                "The following parameters were not understood: %s"
                % ", ".join(unsupported)
            )
            return

        k = {}
        k.update(self.kwargs)
        k.update(kwargs)

        return self.apply_bundle(self.get_bundle_stream(*args, **k), *args, **k)

    def __call__(self, *args, **kwargs):
        return self.run(*args, **kwargs)

    @classmethod
    def as_decorator(cls):
        def _(*args, **kwargs):
            invoked = bool(not args or kwargs)
            if not invoked:
                func, args = args[0], ()
            task_class = kwargs.pop("task_class", cls)

            def wrapper(func):
                return task_class(func, *args, **kwargs)

            return wrapper if invoked else wrapper(func)

        return _


class DeploymentTask(FuselageMixin, tasks.WrappedCallableTask):

    arguments = ["simulate", "loglevel"]

    def apply_bundle(self, bundle, *args, **kwargs):
        uploaded = put(builder.build(bundle), "~/payload.pex", mode=755)
        if uploaded.failed:
            utils.error("Could not upload fuselange bundle to target. Aborting.")
            return

        command = [uploaded[0]]

        if kwargs.get("simulate", "false").lower() in ("true", "y", "yes", "on", "1"):
            command.append("--simulate")

        if "loglevel" in kwargs:
            if kwargs["loglevel"].lower() not in (
                "info",
                "debug",
            ):
                utils.error("Invalid loglevel")
                return
            command.extend(["-v"] * {"info": 0, "debug": 1}[kwargs["loglevel"].lower()])

        try:
            with settings(warn_only=True):
                result = sudo(" ".join(command))

                if not result.return_code in (0, 254):
                    utils.error("Could not apply fuselage blueprint. Aborting.")
                    return

        finally:
            sudo("rm %s" % uploaded[0])


blueprint = DeploymentTask.as_decorator()


class DockerBuildTask(FuselageMixin, tasks.WrappedCallableTask):

    arguments = ["from_image", "tag", "env", "ports", "volumes", "cmd", "maintainer"]

    def apply_bundle(self, bundle, *args, **kwargs):
        from fuselage.docker import DockerBuilder

        d = DockerBuilder(bundle, **kwargs)
        try:
            for data in d.build():
                utils.puts(data, flush=True, end="")
        except RuntimeError as e:
            utils.error(str(e))
            return


docker_container = DockerBuildTask.as_decorator()


loader = Loader(lambda: os.path.join(os.path.dirname(env.real_fabfile), "deployment"))
template = loader.template
static = loader.static
decrypt = loader.decrypt



# Copyright 2011-2014 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Resources for handling the creation and removal of files. These deal with
both the metadata associated with the file (for example owner and permission)
and the contents of the files themselves. """

import os

from fuselage.argument import Boolean, FullPath, String
from fuselage.policy import Policy, Present
from fuselage.resource import Resource
from fuselage.utils import force_str, simple_str


class Line(Resource):

    """
    Ensure that a line is present or missing from a file.

    For example, this will ensure that selinux is disabled::

        Line(
            name="/etc/selinux/config",
            match=r"^SELINUX",
            replace="SELINUX=disabled",
        )
    """

    @property
    def implicit_id(self):
        return force_str(self.name) + ":" + simple_str(self.match)

    name = FullPath()
    """ The full path to the file to perform an operation on. """

    line = String(default="")
    """ The text to insert at the point the expression matches (otherwise at the end of the file). """

    match = String(default="")
    """ The python regular expression to match the line to be updated. """

    linesep = String(default=os.linesep)
    """ The delimiter used to mark a line. \\n on posix. """

    sensitive = Boolean(default=False)


class LineApplyPolicy(Policy):

    """Create a file and populate it's contents if required.

    You must provide a name.

    You may provide one of template, static, or encrypted to act as a file source.
    """

    resource = Line
    name = "apply"
    default = True
    signature = (
        Present("name"),
        Present("line"),
        Present("match"),
    )


class LineRemovePolicy(Policy):

    """
    Ensure lines matching the regular expression are removed from the file.
    """

    resource = Line
    name = "remove"
    signature = (
        Present("name"),
        Present("match"),
    )



# Copyright 2011-2014 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

from fuselage import error, platform, provider, resources
from fuselage.changes import EnsureContents
from fuselage.utils import force_bytes, force_str


class _LineMixin:
    def apply(self):
        if not platform.exists(self.resource.name):
            self.raise_or_log(
                error.PathComponentMissing("File '%s' is missing" % self.resource.name)
            )
            return

        lines = force_str(platform.get(self.resource.name)).splitlines()
        contents = self.resource.linesep.join(line for line in self.filter_lines(lines))

        fc = EnsureContents(
            self.resource.name,
            force_bytes(contents),
            sensitive=self.resource.sensitive,
        )
        self.change(fc)

        return fc.changed


class LineApply(_LineMixin, provider.Provider):

    policies = (resources.line.LineApplyPolicy,)

    def filter_lines(self, lines):
        regexp = re.compile(self.resource.match)
        matched = False
        for line in lines:
            if not matched and regexp.search(line):
                yield self.resource.line
                matched = True
            else:
                yield line

        if not matched:
            yield self.resource.line


class LineRemove(_LineMixin, provider.Provider):

    policies = (resources.line.LineRemovePolicy,)

    def filter_lines(self, lines):
        regexp = re.compile(self.resource.match)
        for line in lines:
            if not regexp.search(line):
                yield line



# Copyright 2011-2014 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from fuselage.argument import Integer, String
from fuselage.policy import NAND, Policy, Present
from fuselage.resource import Resource


class Service(Resource):

    """This represents service startup and shutdown via an init daemon."""

    name = String()
    """ A unique name representing an initd service.

    This would normally match the name as it appears in /etc/init.d.
    """

    priority = Integer(default=99)
    """ Priority of the service within the boot order.

    This attribute will have no effect when using a dependency or event based
    init.d subsystem like upstart or systemd. """

    start = String()
    """ A command that when executed will start the service.

    If not provided, the provider will use the default service start invocation
    for the init.d system in use.
    """

    stop = String()
    """ A command that when executed will start the service.

    If not provided, the provider will use the default service stop invocation
    for the init.d system in use.
    """

    restart = String()
    """ A command that when executed will restart the service.

    If not provided, the provider will use the default service restart invocation
    for the init.d system in use. If it is not possible to automatically determine
    if the restart script is avilable the service will be stopped and started instead.
    """

    reconfig = String()
    """ A command that when executed will make the service reload its
    configuration file. """

    running = String()
    """ A comamnd to execute to determine if a service is running. Should have an exit code of 0 for success. """

    pidfile = String()
    """ Where the service creates its pid file.

    This can be provided instead of a status command as an alternative way of checking if
    a service is running or not.
    """


class ServiceStartPolicy(Policy):

    """Start a service if it isn't running"""

    resource = Service
    name = "start"
    default = True
    signature = (
        Present("name"),
        NAND(Present("running"), Present("pidfile")),
    )


class ServiceStopPolicy(Policy):

    """Stop a service if it is running"""

    resource = Service
    name = "stop"
    signature = (
        Present("name"),
        NAND(Present("running"), Present("pidfile")),
    )


class ServiceRestartPolicy(Policy):

    """Restart a service

    If a service isn't running it will just be started instead.
    """

    resource = Service
    name = "restart"
    signature = (
        Present("name"),
        NAND(Present("running"), Present("pidfile")),
    )


class ServiceReloadPolicy(Policy):

    """Get the service to reload its configuration

    If a service isn't running it will just be started instead.
    """

    resource = Service
    name = "reconfig"
    signature = (
        Present("name"),
        NAND(Present("running"), Present("pidfile")),
    )



# Copyright 2011-2014 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from fuselage import error, platform, provider, resources
from fuselage.changes import EnsureDirectory, ShellCommand


class Svn(provider.Provider):

    policies = (resources.checkout.CheckoutSyncPolicy,)

    @classmethod
    def isvalid(self, policy, resource):
        identities = [
            "svn",
            "subversion",
        ]

        return resource.scm.lower() in identities

    @property
    def url(self):
        repository = self.resource.repository
        if self.resource.tag:
            return repository + "/tags/" + self.resource.tag
        return repository + "/" + self.resource.branch

    def apply(self):
        if not platform.exists("/usr/bin/svn"):
            self.raise_or_log(
                error.MissingDependency(
                    "'/usr/bin/svn' is not available; update your configuration to install subversion?"
                )
            )

        self.change(
            EnsureDirectory(
                self.resource.name, self.resource.user, self.resource.group, 0o755
            )
        )

        if not platform.exists(os.path.join(self.resource.name, ".svn")):
            self.svn("co", self.url, self.resource.name)
            return True

        changed = False

        info = self.info(self.resource.name)
        repo_info = self.info(self.url)

        # If the 'Repository Root' is different between the checkout and the
        # repo, switch --relocated
        old_repo_root = info["Repository Root"]
        new_repo_root = repo_info["Repository Root"]
        if old_repo_root != new_repo_root:
            self.changelog.info(
                "Switching repository root from '%s' to '%s'"
                % (old_repo_root, new_repo_root)
            )
            self.svn(
                "switch", "--relocate", old_repo_root, new_repo_root, self.resource.name
            )
            changed = True

        # If we have changed branch, switch
        old_url = info["URL"]
        new_url = repo_info["URL"]
        if old_url != new_url:
            self.changelog.info(f"Switching branch from '{old_url}' to '{new_url}'")
            self.svn("switch", new_url, self.resource.name)
            changed = True

        # If we have changed revision, svn up
        # FIXME: Eventually we might want revision to be specified in the
        # resource?
        current_rev = info["Last Changed Rev"]
        target_rev = repo_info["Last Changed Rev"]
        if current_rev != target_rev:
            self.changelog.info(
                f"Switching revision from {current_rev} to {target_rev}"
            )
            self.svn("up", "-r", target_rev, self.resource.name)
            changed = True

        return changed

    def action_export(self):
        if platform.exists(self.resource.name):
            return
        self.svn("export", self.url, self.resource.name)

    def get_svn_args(self, action, *args, **kwargs):
        command = ["svn"]

        if kwargs.get("quiet", False):
            command.append("--quiet")

        command.extend([action, "--non-interactive"])

        scm_username = self.resource.scm_username
        scm_password = self.resource.scm_password
        if scm_username:
            command.extend(["--username", self.resource.scm_username])
        if scm_password:
            command.extend(["--password", self.resource.scm_password])
        if scm_username or scm_password:
            command.append("--no-auth-cache")

        for arg in args:
            command.append(arg)

        return command

    def info(self, uri):
        command = self.get_svn_args("info", uri)
        stdout, stderr = platform.check_call(command)
        return dict(x.split(": ") for x in stdout.split("\n") if x)

    def svn(self, action, *args, **kwargs):
        command = self.get_svn_args(action, *args, **kwargs)
        sc = ShellCommand(command, user=self.resource.user)
        self.change(sc)



# Copyright 2014 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import io
import json
import tarfile

from fuselage.builder import build

try:
    import docker
except ImportError:
    docker = None


class DockerBuilder:
    def __init__(
        self,
        bundle,
        from_image="ubuntu",
        tag=None,
        env=None,
        volumes=None,
        ports=None,
        cmd=None,
        maintainer=None,
    ):
        self.bundle = bundle
        self.from_image = from_image
        self.tag = tag
        self.env = env or {}
        self.volumes = volumes or []
        self.ports = ports or []
        self.cmd = cmd
        self.maintainer = maintainer

    def get_dockerfile(self):
        df = [
            "# This Dockerfile was automatically generated by fuselage",
        ]

        if self.maintainer:
            df.append("MAINTAINER %s" % self.maintainer)

        df.append("FROM %s" % self.from_image or "ubuntu")
        df.append("")

        for k, v in self.env.items():
            df.append(f"ENV {k} {v}")

        if self.volumes:
            df.append("VOLUME %s" % json.dumps(self.volumes))

        if self.ports:
            df.append("EXPOSE %s" % " ".join(str(p) for p in self.ports))

        if self.env or self.volumes or self.ports:
            df.append("")

        df.extend(
            [
                "RUN if [ -f /usr/bin/apt-get ]; then apt-get update && apt-get install python -y; fi",
                "RUN if [ -f /usr/bin/yum ]; then yum install python -y; fi",
                "",
                "ADD payload.pex /payload.pex",
                "RUN /payload.pex",
                "RUN rm /payload.pex",
            ]
        )

        if self.cmd:
            df.append("")
            df.append("CMD %s" % json.dumps(self.cmd))

        return "\n".join(df)

    def build(self):
        tar_buffer = io.BytesIO()
        tar = tarfile.open(mode="w:gz", fileobj=tar_buffer)

        def add(name, buf, mode=0o644):
            ti = tarfile.TarInfo(name=name)
            ti.size = len(buf.buf)
            ti.mode = mode
            tar.addfile(tarinfo=ti, fileobj=buf)

        add("payload.pex", build(self.bundle), mode=0o755)
        add("Dockerfile", io.BytesIO(self.get_dockerfile()))

        tar.close()
        tar_buffer.seek(0)

        c = docker.Client(
            base_url="unix://var/run/docker.sock",
            version="1.12",
            timeout=10,
        )

        build_output = c.build(
            fileobj=tar_buffer,
            custom_context=True,
            stream=True,
            rm=True,
            tag=self.tag,
        )

        status = None
        for data in build_output:
            data = json.loads(data)
            if "status" in data:
                # {u'status': u'Downloading', u'progressDetail': {u'current': 3239, u'start': 141059980, u'total': 133813}, u'id': u'2124c4204a05', u'progress': u'[=>] 32.32 kB/1.338 MB 29s'}
                if status != data["status"]:
                    if status:
                        yield "\n"
                    status = data["status"]
                    yield status
                else:
                    yield "."
            else:
                if status:
                    yield "\n"
                    status = None

                if "stream" in data:
                    yield data["stream"]
                elif "errorDetail" in data:
                    raise RuntimeError(data["errorDetail"]["message"])



# Copyright 2011-2014 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Core classes for providers """

from abc import ABCMeta, abstractmethod
import logging

from fuselage import log, policy


class ProviderType(ABCMeta):

    """Registers the provider with the resource which it provides"""

    def __new__(meta, class_name, bases, new_attrs):
        cls = super().__new__(meta, class_name, bases, new_attrs)
        for p in getattr(cls, "policies", []):
            p.providers.append(cls)
        return cls


class Provider(metaclass=ProviderType):

    # every provider should have a name
    name = None
    # in your class, specify which policies you provide an implementation for
    # these policies should all be for the same resource
    policies = []

    def __init__(self, resource, runner):
        self.resource = resource
        self.runner = runner

        logger = logging.getLogger(self.__module__)
        self.logger = log.LoggerAdapter(logger, {"fuselage.resource": resource.id})

        # logger = logging.getLogger("fuselage.audit")
        self.changelog = log.LoggerAdapter(
            self.logger,
            {
                # "resource": resource.id,
                "fuselage.changelog": True,
            },
        )

    @property
    def simulate(self):
        return self.runner.simulate

    def raise_or_log(self, exc):
        if not self.simulate:
            raise exc
        self.logger.warning(exc)

    def change(self, change):
        return change.apply(self)

    @classmethod
    def isvalid(self, policy, resource):
        """Returns True if this provider is valid for the specified resource,
        within the context of the provided yay structure. This returns True by
        default. If you want your provider to be more discriminating, then
        make it so. In particular if you want two providers for a policy, then
        only one of those providers may return True from this method."""
        return True

    @abstractmethod
    def apply(self, shell):
        """Execute this provider using the supplied shell object. This base
        method must be overridden. This should return True if the provider
        changed anything, or False if it did not need to change anything (i.e.
        the Resource was already in the state the policy ensures)."""
        return False


class NullProvider(Provider):
    policies = [policy.NullPolicy]

    def apply(self, context, output):
        pass



# Copyright 2011-2014 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from fuselage.argument import Boolean, Integer, String
from fuselage.policy import Absent, Policy, Present
from fuselage.resource import Resource


class Group(Resource):

    """A resource representing a unix group stored in the /etc/group file.
    groupadd and groupmod are used to actually make modifications.

    For example::

        Group:
            name: zope
            system: true
    """

    name = String()
    """ The name of the unix group. """

    gid = Integer()
    """ The group ID associated with the group. If this is not specified one will be chosen. """

    system = Boolean(default=False)
    """ Whether or not this is a system group - i.e. the new group id will be
    taken from the system group id list. """

    password = String()
    """ The password for the group, if required """


class GroupApplyPolicy(Policy):

    """Create the group, or ensure it has the specified attributes."""

    resource = Group
    name = "apply"
    default = True
    signature = (Present("name"),)


class GroupRemovePolicy(Policy):

    """Remove an existing group if it still exists.

    You should only specify the name of the group when using this policy."""

    resource = Group
    name = "remove"
    default = False
    signature = (
        Present("name"),
        Absent("gid"),
        Absent("system"),
        Absent("password"),
    )



# Copyright 2011-2014 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import stat

from fuselage import error, platform
from fuselage.changes import base

from .execute import ShellCommand


class AttributeChanger(base.Change):

    """Make the changes required to a file's attributes"""

    def __init__(self, filename, user=None, group=None, mode=None):
        self.filename = filename
        self.user = user
        self.group = group
        self.mode = mode
        self.changed = False

    def apply(self, context):
        """Apply the changes"""

        if not platform.getpwnam or not platform.getgrnam:
            return self

        uid = None
        gid = None
        mode = None

        if platform.exists(self.filename):
            st = platform.stat(self.filename)
            uid = st.st_uid
            gid = st.st_gid
            mode = stat.S_IMODE(st.st_mode)

        if self.user is not None:
            try:
                owner = platform.getpwnam(self.user)
            except KeyError:
                context.raise_or_log(
                    error.InvalidUser("User '%s' not found" % self.user)
                )
                owner = None

            if not owner or owner.pw_uid != uid:
                context.change(ShellCommand(["chown", self.user, self.filename]))
                self.changed = True

        if self.group is not None:
            try:
                group = platform.getgrnam(self.group)
            except KeyError:
                context.raise_or_log(
                    error.InvalidGroup("No such group '%s'" % self.group)
                )
                group = None

            if not group or group.gr_gid != gid:
                context.change(ShellCommand(["chgrp", self.group, self.filename]))
                self.changed = True

        if self.mode is not None and mode is not None:
            if mode != self.mode:
                context.change(ShellCommand(["chmod", "%o" % self.mode, self.filename]))

                # Clear the user and group bits
                # We don't need to set them as chmod will *set* this bits with an octal
                # but won't clear them without a symbolic mode
                if mode & stat.S_ISGID and not self.mode & stat.S_ISGID:
                    context.change(ShellCommand(["chmod", "g-s", self.filename]))
                if mode & stat.S_ISUID and not self.mode & stat.S_ISUID:
                    context.change(ShellCommand(["chmod", "u-s", self.filename]))

                self.changed = True

        return self



# Copyright 2011-2014 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import random
import sys

from fuselage import error
from fuselage.utils import force_str


class Argument:

    """Stores the argument value on the instance object. It's a bit fugly,
    neater ways of doing this that do not involve passing extra arguments to
    Argument are welcome."""

    argument_id = 0
    default = None

    def __init__(self, **kwargs):
        self.default = kwargs.pop("default", self.default)
        self.__doc__ = kwargs.pop("help", None)
        self.arg_id = "argument_%d" % Argument.argument_id
        Argument.argument_id += 1

    def __get__(self, instance, owner):
        if instance is None:
            return self
        return self.get(instance)

    def get(self, instance):
        if instance is None:
            return self
        elif self.present(instance):
            return self.get_raw(instance)
        else:
            return self.get_default(instance)

    def get_default(self, instance):
        if callable(self.default):
            return self.default(instance)
        return self.default

    def get_raw(self, instance):
        return getattr(instance, self.arg_id)

    def present(self, instance):
        return hasattr(instance, self.arg_id)

    def serialize(self, instance, builder=None):
        return self.get(instance)

    def clean(self, instance, value):
        return value

    def __set__(self, instance, value):
        """Set the property. The value will be a UTF-8 encoded string read from the yaml source file."""
        value = self.clean(instance, value)
        self.save(instance, value)

    def save(self, instance, value):
        setattr(instance, self.arg_id, value)


class Boolean(Argument):

    """Represents a boolean. "1", "yes", "on" and "true" are all considered
    to be True boolean values. Anything else is False."""

    def clean(self, instance, value):
        if isinstance(value, str):
            if value.lower() in ("1", "yes", "on", "true"):
                value = True
            else:
                value = False
        else:
            value = bool(value)
        return value

    @classmethod
    def _generate_valid(self):
        return random.choice([True, False])


class String(Argument):

    """Represents a string."""

    def clean(self, instance, value):
        if value is not None:
            value = force_str(value)
        return value

    @classmethod
    def _generate_valid(self):
        glyphs = []
        for i in range(random.randint(0, 1024)):
            glyphs.append(
                random.choice(
                    [
                        "\u2603",
                        "\U0001F42D",
                        "/",
                        "1",
                        "+",
                        "-",
                        ".",
                        "`",
                        '"',
                        "'",
                        "a",
                        "b",
                        "c",
                        "d",
                        "E",
                        "F",
                    ]
                )
            )
        return "".join(glyphs)


class FullPath(String):

    """Represents a full path on the filesystem. This should start with a
    '/'."""

    def clean(self, instance, value):
        return super().clean(instance, value)

    @classmethod
    def _generate_valid(self):
        return "/" + super()._generate_valid()


class Integer(Argument):

    """Represents an integer argument taken from the source file. This can
    throw an :py:exc:error.ParseError if the passed in value cannot represent
    a base-10 integer."""

    def clean(self, instance, value):
        if not isinstance(value, int):
            try:
                value = int(value)
            except ValueError:
                raise error.ParseError("%s is not an integer" % value)
        return value

    @classmethod
    def _generate_valid(self):
        return random.randint(0, sys.maxsize)


class Octal(Integer):

    """An octal integer.  This is specifically used for file permission modes."""

    def clean(self, instance, value):
        if not isinstance(value, int):
            value = int(value, 8)
        return value


class Dict(Argument):
    @classmethod
    def _generate_valid(self):
        return {}


class List(Argument):
    @classmethod
    def _generate_valid(self):
        return []


class File(Argument):
    def _generate_valid(self):
        return "/tmp/foo"

    def serialize(self, instance, builder=None):
        assert builder
        if not self.present(instance):
            return self.get_default(instance)

        filename = self.get_raw(instance)
        with open(filename, "rb") as fp:
            return builder.add_resource_blob(fp.read())


class PolicyTrigger:
    def __init__(self, on):
        self.on = on

    def bind(self, resources, target):
        try:
            resource = resources.get_resource_by_name(self.on)
        except Exception:
            try:
                resource = resources[self.on]
            except KeyError:
                raise error.BindingError(
                    f"Cannot bind {target!r} to missing resource named '{self.on}'"
                )

        resource.register_observer("*", target, "*")
        return resource


class SubscriptionArgument(Argument):

    """Parses the policy: argument for resources, including triggers etc."""

    def clean(self, instance, value):
        triggers = []
        for resource in value:
            if isinstance(resource, PolicyTrigger):
                trigger = resource
            else:
                trigger = PolicyTrigger(resource)
            triggers.append(trigger)
        return triggers

    def serialize(self, instance, builder=None):
        if not self.present(instance):
            return
        value = self.get_raw(instance)
        policy = []
        for t in value:
            policy.append(t.on)
        return policy

    def _generate_valid(self):
        return []


class PolicyArgument(Argument):

    """Parses the policy: argument for resources, including triggers etc."""

    def clean(self, instance, value):
        """Set either a default policy or a set of triggers on the policy collection"""
        try:
            value = instance.policies[value](instance)
        except KeyError:
            raise error.ParseError(f"'{value}' is not a valid policy for this resource")
        return value

    def get_default(self, instance):
        return instance.policies.default()(instance)

    def serialize(self, instance, builder=None):
        if not self.present(instance):
            return
        return self.get_raw(instance).name

    def _generate_valid(self):
        return "invalid-policy"



# Copyright 2014 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import binascii
import os

from libcloud.compute import deployment

from fuselage import builder
from fuselage.bundle import ResourceBundle


class FuselageDeployment(deployment.Deployment):
    def __init__(self, bundle=None, resources=None):
        self.bundle = bundle

        if not self.bundle:
            self.bundle = ResourceBundle()

        if resources:
            self.bundle.extend(resources)

    def run(self, node, client):
        random_string = binascii.hexlify(os.urandom(4)).decode("ascii")
        name = "fuselage_%s" % (random_string)

        bu = builder.build(self.bundle)

        file_path = client.put(path=name, chmod=0o755, contents=bu.getvalue())
        self.stdout, self.stderr, self.exit_status = client.run(file_path)
        client.delete(name)
        return node



# Copyright 2011-2014 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import stat

from fuselage import error, platform, provider, resources
from fuselage.changes import ShellCommand


class Link(provider.Provider):

    policies = (resources.link.LinkAppliedPolicy,)

    def _get_owner(self):
        """Return the uid for the resource owner, or None if no owner is
        specified."""
        owner = self.resource.owner
        if owner:
            try:
                return platform.getpwnam(owner).pw_uid
            except KeyError:
                raise error.InvalidUser()

    def _get_group(self):
        """Return the gid for the resource group, or None if no group is
        specified."""
        group = self.resource.group
        if group:
            try:
                return platform.getgrnam(group).gr_gid
            except KeyError:
                raise error.InvalidGroup()

    def _stat(self):
        """Extract stat information for the resource."""
        st = platform.lstat(self.resource.name)
        uid = st.st_uid
        gid = st.st_gid
        mode = stat.S_IMODE(st.st_mode)
        return uid, gid, mode

    def apply(self):
        changed = False
        name = self.resource.name
        to = self.resource.to
        uid = None
        gid = None
        mode = None
        isalink = False

        if not platform.exists(to):
            self.raise_or_log(
                error.DanglingSymlink("Destination of symlink %r does not exist" % to)
            )

        owner = self._get_owner()
        group = self._get_group()

        try:
            linkto = platform.readlink(name)
            isalink = True
        except OSError:
            isalink = False

        if not isalink or linkto != to:
            if platform.lexists(name):
                self.change(ShellCommand(["/bin/rm", "-rf", self.resource.name]))

            self.change(
                ShellCommand(["/bin/ln", "-s", self.resource.to, self.resource.name])
            )
            changed = True

        try:
            linkto = platform.readlink(name)
            isalink = True
        except OSError:
            isalink = False

        if not isalink:
            if not self.simulate:
                raise error.OperationFailed("Did not create expected symbolic link")
            return changed

        uid, gid, mode = self._stat()

        if owner and owner != uid:
            self.change(
                ShellCommand(
                    ["/bin/chown", "-h", self.resource.owner, self.resource.name]
                )
            )
            changed = True

        if group and group != gid:
            self.change(
                ShellCommand(
                    ["/bin/chgrp", "-h", self.resource.group, self.resource.name]
                )
            )
            changed = True

        return changed


class RemoveLink(provider.Provider):

    policies = (resources.link.LinkRemovedPolicy,)

    def apply(self):
        name = self.resource.name

        if platform.lexists(name):
            if not platform.islink(name):
                raise error.InvalidProvider(
                    f"{self!r}: {name} exists and is not a link"
                )
            self.change(ShellCommand(["/bin/rm", self.resource.name]))
            return True
        return False



# Copyright 2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from . import (
    apt,
    directory,
    execute,
    files,
    git,
    group,
    line,
    link,
    mercurial,
    mount,
    mounted,
    patch,
    service,
    subversion,
    user,
    yum,
)

__all__ = [
    "apt",
    "directory",
    "execute",
    "files",
    "git",
    "group",
    "line",
    "link",
    "mercurial",
    "mount",
    "mounted",
    "service",
    "subversion",
    "patch",
    "user",
    "yum",
]



# Copyright 2011-2014 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from fuselage import error, platform, provider, resources
from fuselage.changes import ShellCommand


class _ServiceMixin:
    features = [
        "restart",
    ]

    def status(self):
        if self.resource.running:
            self.logger.debug(
                "Running %r to determine if already running" % self.resource.running
            )
            try:
                platform.check_call(self.resource.running)
            except error.SystemError as e:
                self.logger.debug(
                    "Got exit code %d. Assuming not running." % (e.returncode,)
                )
                return "not-running"
            else:
                self.logger.debug("Got exit code 0. Assuming running.")
                return "running"

        if not self.resource.pidfile:
            self.logger.debug(
                "Neither 'pidfile' nor 'running' option is set. Cannot determine service state."
            )
            return "unknown"

        self.logger.debug(
            f"Using pidfile {self.resource.pidfile!r} to determine service state"
        )

        if not platform.exists(self.resource.pidfile):
            self.logger.debug("The pidfile does not exist, service not running")
            return "not-running"

        try:
            pid = int(platform.get(self.resource.pidfile).strip())
        except Exception:
            self.logger.debug(
                "The pidfile could not be understood (should just contain a single int). Cannot determine service state."
            )
            return "unknown"

        # if platform.exists("/proc/%s" % pid):
        #     return "running"

        try:
            platform.check_call(["kill", "-0", str(pid)])
            self.logger.debug("Service is running.")
            return "running"
        except error.SystemError:
            self.logger.debug(
                "Unable to kill(0) pid %d - service is not running." % pid
            )
            return "not-running"

    def do(self, action):
        try:
            self.change(ShellCommand(self.get_command(action)))
        except error.SystemError as exc:
            raise error.CommandError(
                "%s failed with return code %d" % (action, exc.returncode)
            )

    def ensure_enabled(self):
        pass

    def ensure_disabled(self):
        pass

    def get_command(self, action):
        return getattr(self.resource, action)


class Start(_ServiceMixin, provider.Provider):

    policies = (resources.service.ServiceStartPolicy,)

    def apply(self):
        changed = self.ensure_enabled()

        if self.status() == "running":
            return changed

        self.do("start")

        return True


class Stop(_ServiceMixin, provider.Provider):

    policies = (resources.service.ServiceStopPolicy,)

    def apply(self):
        changed = self.ensure_disabled()

        if self.status() == "not-running":
            return changed

        self.do("stop")

        return True


class Restart(_ServiceMixin, provider.Provider):

    policies = (resources.service.ServiceRestartPolicy,)

    def apply(self):
        self.ensure_enabled()

        if self.status() == "not-running":
            self.do("start")
            return True

        if "restart" in self.features:
            self.do("restart")
        else:
            self.do("stop")
            self.do("start")

        return True



# Copyright 2014 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
import unicodedata


def force_str(s):
    if isinstance(s, str):
        return s
    elif isinstance(s, bytes):
        return s.decode("utf-8")
    raise ValueError("Not a string")


def force_unicode(s):
    if isinstance(s, bytes):
        return s.decode("utf-8")
    elif isinstance(s, str):
        return s
    raise ValueError("Not a string")


def force_bytes(s):
    if isinstance(s, bytes):
        return s
    elif isinstance(s, str):
        return s.encode("utf-8")
    raise ValueError("Not a string")


def simple_str(s):
    """'Normalize' a string - lowercase, strip unicode, replace whitespace, etc"""
    s = force_str(
        unicodedata.normalize("NFKD", force_unicode(s)).encode("ascii", "ignore")
    )
    s = s.lower().strip()
    s = re.sub(r"[^\w\s-]", "", s)
    s = re.sub(r"[-\s]+", "-", s)
    return s



# Copyright 2011-2014 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import re

from fuselage import error, platform, provider, resources
from fuselage.changes import EnsureDirectory, ShellCommand

log = logging.getLogger(__name__)


class Git(provider.Provider):

    policies = (resources.checkout.CheckoutSyncPolicy,)

    REMOTE_NAME = "origin"

    @classmethod
    def isvalid(self, policy, resource):
        return resource.scm and resource.scm.lower() == "git"

    def get_git_command(self, action, *args):
        command = [
            "git",
            # "--git-dir=%s" % os.path.join(self.resource.name, ".git"),
            # "--work-tree=%s" % self.resource.name,
            "--no-pager",
            action,
        ]

        command.extend(list(args))
        return command

    def info(self, action, *args):
        try:
            stdout, stderr = platform.check_call(
                self.get_git_command(action, *args),
                user=self.resource.user,
                cwd=self.resource.name,
            )
            returncode = 0
        except error.SystemError as e:
            returncode, stdout, stderr = e.returncode, e.stdout, e.stderr

        return returncode, stdout, stderr

    def action(self, action, *args):
        self.change(
            ShellCommand(
                self.get_git_command(action, *args),
                user=self.resource.user,
                cwd=self.resource.name,
            )
        )

    def action_clone(self):
        """Adds resource.repository as a remote, but unlike a
        typical clone, does not check it out

        """
        self.change(
            EnsureDirectory(
                self.resource.name, self.resource.user, self.resource.group, 0o755
            )
        )

        try:
            self.action("init", self.resource.name)
        except error.SystemError:
            raise error.CheckoutError("Cannot initialise local repository.")

        self.action_set_remote()

    def action_set_remote(self):
        try:
            self.action("remote", "add", self.REMOTE_NAME, self.resource.repository)
        except error.SystemError:
            raise error.CheckoutError("Could not set the remote repository.")

    def action_update_remote(self):
        # Determine if the remote repository has changed
        remote_re = re.compile(self.REMOTE_NAME + r"\t(.*) \(.*\)\n")
        rv, stdout, stderr = self.info("remote", "-v")
        remote = remote_re.search(stdout)
        if remote:
            if not self.resource.repository == remote.group(1):
                log.info("The remote repository has changed.")
                try:
                    self.action("remote", "rm", self.REMOTE_NAME)
                except error.SystemError:
                    raise error.CheckoutError(
                        "Could not delete remote '%s'" % self.REMOTE_NAME
                    )
                self.action_set_remote()
                return True
        else:
            raise error.CheckoutError("Cannot determine repository remote.")

        return False

    def checkout_needed(self):
        # Determine which SHA is currently checked out.
        if platform.exists(os.path.join(self.resource.name, ".git")):
            try:
                rv, stdout, stderr = self.info("rev-parse", "--verify", "HEAD")
            except error.SystemError:
                head_sha = "0" * 40
            else:
                head_sha = stdout[:40]
                log.info("Current HEAD sha: %s" % head_sha)
        else:
            head_sha = "0" * 40

        try:
            stdout, stderr = platform.check_call(
                command=["git", "ls-remote", self.resource.repository],
                user=self.resource.user,
                cwd="/tmp",
            )
        except error.SystemError:
            raise error.CheckoutError("Could not query the remote repository")

        r = re.compile("([0-9a-f]{40})\t(.*)\n")
        refs_to_shas = {b: a for (a, b) in r.findall(stdout)}

        # Revision takes precedent over branch

        revision = self.resource.revision
        tag = self.resource.tag
        branch = self.resource.branch

        if revision:
            newref = revision
            if newref != head_sha:
                return newref

        elif tag:
            as_tag = "refs/tags/%s" % tag
            if as_tag not in refs_to_shas.keys():
                raise error.CheckoutError("Cannot find a tag called '%s'" % tag)

            annotated_tag = as_tag + "^{}"
            if annotated_tag in refs_to_shas.keys():
                as_tag = annotated_tag
            newref = tag
            if head_sha != refs_to_shas.get(as_tag):
                return newref

        elif branch:
            as_branch = "refs/heads/%s" % branch
            if as_branch not in refs_to_shas.keys():
                raise error.CheckoutError("Cannot find a branch called '%s'" % branch)
            newref = f"remotes/{self.REMOTE_NAME}/{branch}"
            if head_sha != refs_to_shas.get(as_branch):
                return newref
        else:
            raise error.CheckoutError(
                "You must specify either a revision, tag or branch"
            )

    def action_checkout(self, newref):
        try:
            self.action("fetch", self.REMOTE_NAME)
        except error.SystemError:
            raise error.CheckoutError("Could not fetch '%s'" % self.resource.repository)

        try:
            self.action("checkout", newref)
        except error.SystemError:
            raise error.CheckoutError("Could not check out '%s'" % newref)

    def apply(self):
        if not platform.exists("/usr/bin/git"):
            self.raise_or_log(
                error.MissingDependency(
                    "'/usr/bin/git' is not available; update your configuration to install git?"
                )
            )
            return

        # If necessary, clone the repository
        if not platform.exists(os.path.join(self.resource.name, ".git")):
            self.action_clone()
            changed = True
        else:
            changed = self.action_update_remote()

        newref = self.checkout_needed()
        if newref:
            self.action_checkout(newref)

        return changed or newref



# Copyright 2011-2014 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from fuselage import error, platform, provider, resources
from fuselage.changes import EnsureFile
from fuselage.utils import force_bytes


class Patch(provider.Provider):

    """Provides file creation using templates or static files."""

    policies = (resources.patch.PatchApplyPolicy,)

    def check_path(self, directory):
        if platform.isdir(directory):
            return
        frags = directory.split("/")
        path = "/"
        for i in frags:
            path = os.path.join(path, i)
            if not platform.exists(path):
                self.raise_or_log(error.PathComponentMissing(path))
            elif not platform.isdir(path):
                raise error.PathComponentNotDirectory(path)

    def apply_patch(self):
        patch = self.resource.patch
        try:
            stdout, stderr = platform.check_call(
                command=[
                    "patch",
                    "-t",
                    "--dry-run",
                    "-N",
                    "--silent",
                    "-r",
                    "-",
                    "-o",
                    "-",
                    self.resource.source,
                    "-",
                ],
                stdin=patch,
            )
        except error.SystemError as e:
            self.logger.error("Patch does not apply cleanly")
            self.logger.error("Patch file used was %s" % self.resource.patch)
            self.logger.error("File to patch was %s" % self.resource.source)

            self.logger.error("")
            self.logger.error("Reported error was:")
            map(self.logger.error, e.stderr.split("\n"))

            raise error.CommandError("Unable to apply patch")

        return stdout

    def apply(self):
        name = self.resource.name

        self.check_path(os.path.dirname(name))

        contents = force_bytes(self.apply_patch())

        fc = EnsureFile(
            name,
            contents,
            self.resource.owner,
            self.resource.group,
            self.resource.mode,
            sensitive=self.resource.sensitive,
        )
        self.change(fc)

        return fc.changed



# Copyright 2011-2014 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Resources dealing with filesystem objects other than files. """

from fuselage.argument import Boolean, FullPath, Octal, String
from fuselage.defaults import get_default_group, get_default_user
from fuselage.policy import Absent, Policy, Present
from fuselage.resource import Resource


class Directory(Resource):

    """A directory on disk. Directories have limited metadata, so this
    resource is quite limited.

    For example::

        Directory:
          name: /var/local/data
          owner: root
          group: root
          mode: 0755

    """

    name = FullPath()
    """ The full path to the directory on disk """

    owner = String(default=lambda r: get_default_user())
    """ The unix username who should own this directory, by default this is 'root' """

    group = String(default=lambda r: get_default_group(r.owner))
    """ The unix group who should own this directory, by default this is 'root' """

    mode = Octal(default=0o755)
    """ The octal mode that represents this directory's permissions, by default this is '755'. """

    parents = Boolean(default=False)
    """ Create parent directories as needed, using the same ownership and
    permissions, this is False by default. """


class DirectoryAppliedPolicy(Policy):

    """Ensure a directory exists and matches the specification provided
    by the resource."""

    resource = Directory
    name = "apply"
    default = True
    signature = (Present("name"),)


class DirectoryRemovedPolicy(Policy):

    """If a directory described by this resource exists then remove it.

    This isn't recursive, if you want to remove a directory and all its contents
    use `remove-recursive`.

    You should only provide the path to the directory when using this policy.
    Other fields are not needed.
    """

    resource = Directory
    name = "remove"
    default = False
    signature = (
        Present("name"),
        Absent("owner"),
        Absent("group"),
        Absent("mode"),
    )


class DirectoryRemovedRecursivePolicy(Policy):

    """If a directory described by this resource exists then remove it and
    its children.

    You should only provide the path to the directory when using this policy.
    Other fields are not needed.
    """

    resource = Directory
    name = "remove-recursive"
    default = False
    signature = (
        Present("name"),
        Absent("owner"),
        Absent("group"),
        Absent("mode"),
    )



# Copyright 2011-2014 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Resources dealing with filesystem objects other than files. """

from fuselage.argument import FullPath, Integer, Octal, String
from fuselage.policy import Absent, Policy, Present
from fuselage.resource import Resource


class Special(Resource):

    """A special file, as created by mknod."""

    name = FullPath()
    """ The full path to the special file on disk. """

    owner = String(default="root")
    """ The unix user who should own this special file. """

    group = String(default="root")
    """ The unix group who should own this special file. """

    mode = Octal(default=0o644)
    """ The octal representation of the permissions for this special file. """

    type = String(default="fifo")
    """ One of the following strings:

      block
        create a block (buffered) special file
      character
        create a character (unbuffered) special file
      fifo
        create a fifo

    It defaults to fifo
    """

    major = Integer()
    """ The major number for the special file. If the type of the special file
    is block or character, then this must be specified. """

    minor = Integer()
    """ The minor number for the special file. If the type of the special file
    is block or character, then this must be specified. """


class SpecialAppliedPolicy(Policy):

    """Ensure a block or character special file exists"""

    name = "apply"
    default = True
    signature = (
        Present("name"),
        Present("owner"),
        Present("group"),
        Present("mode"),
        Present("type"),
        Present("major"),
        Present("minor"),
    )


class SpecialRemovedPolicy(Policy):

    """If the special file specified exists, remove it.

    You should only specify the special file to remove, the other fields are
    not needed"""

    name = "remove"
    default = False
    signature = (
        Present("name"),
        Absent("owner"),
        Absent("group"),
        Absent("mode"),
        Absent("type"),
        Absent("major"),
        Absent("minor"),
    )



# Copyright 2013-2014 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .bundle import ResourceBundle
from .resources import (
    Checkout,
    Directory,
    Execute,
    File,
    Group,
    Line,
    Link,
    Mount,
    Package,
    Patch,
    Service,
    Special,
    User,
)

__all__ = [
    "ResourceBundle",
    "Checkout",
    "Directory",
    "Execute",
    "File",
    "Group",
    "Line",
    "Link",
    "Mount",
    "Package",
    "Patch",
    "Service",
    "Special",
    "User",
]



# Copyright 2012-2014 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import optparse
import os
import pkgutil
import sys

from fuselage import bundle, error, event, log, platform
from fuselage.error import NothingChanged
from fuselage.utils import force_str

logger = logging.getLogger(__name__)


class Runner:

    state_path = "/var/run/yaybu"

    def __init__(
        self,
        resources,
        resume=False,
        no_resume=False,
        no_changes_ok=False,
        simulate=False,
        verbosity=logging.INFO,
        state_path=None,
    ):
        if resume and no_resume:
            raise error.ParseError("'resume' and 'no_resume' cannot both be True")

        if state_path is not None:
            self.state_path = state_path

        self.resources = resources
        self.resume = resume
        self.no_resume = no_resume
        self.no_changes_ok = no_changes_ok
        self.simulate = simulate
        self.verbosity = verbosity

        self.state = event.EventState(
            save_file=os.path.join(self.state_path, "events.saved"),
            simulate=self.simulate,
        )

    @classmethod
    def get_resources(cls):
        return bundle.ResourceBundle()

    @classmethod
    def setup_from_cmdline(cls, argv=sys.argv, resources=None):
        p = optparse.OptionParser()
        p.add_option("--state", default=None)
        p.add_option("-s", "--simulate", action="store_true", default=False)
        p.add_option("--resume", action="store_true", default=False)
        p.add_option("--no-resume", action="store_true", default=False)
        p.add_option("--no-changes-ok", action="store_true", default=False)
        p.add_option("-v", "--verbose", action="count", default=0)
        p.add_option("-q", "--quiet", action="count", default=0)
        opts, args = p.parse_args(argv)

        return cls(
            resources or cls.get_resources(),
            resume=opts.resume,
            no_resume=opts.no_resume,
            no_changes_ok=opts.no_changes_ok,
            simulate=opts.simulate,
            verbosity=logging.INFO - (10 * (opts.verbose - opts.quiet)),
            state_path=opts.state,
        )

    def run(self):
        log.configure(verbosity=self.verbosity, force=True)

        logger.debug("Runner started")
        logger.debug("Created runner with %d resources" % len(self.resources))

        if not self.simulate and not platform.exists(self.state_path):
            logger.debug("Creating state directories at %r" % self.state_path)
            platform.makedirs(self.state_path)

        self.state.open()

        try:
            changed = self.resources.apply(self)
        except NothingChanged:
            if not self.no_changes_ok:
                raise
            changed = []

        # FIXME: Do we get here if no change has occured??
        self.state.success()

        return changed


class BundledRunner(Runner):
    @classmethod
    def get_resources(self):
        loader = pkgutil.get_loader("fuselage")
        try:
            resources_json = loader.get_data("resources.json").decode("ascii")
        except OSError:
            raise error.ParseError("Bundle is missing resources.json")
        b = bundle.ResourceBundle()
        b.loads(force_str(resources_json))
        return b



# Copyright 2013-2014 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from fuselage.argument import Boolean, File, FullPath, Integer, Octal, String
from fuselage.policy import Policy, Present
from fuselage.resource import Resource


class Patch(Resource):

    """
    A provider for this resource will copy a source file from the target system
    and apply a patch to it.

    For example::

        - Patch:
            target: /output
            source: /output
            patch: localfile

    """

    name = FullPath()
    """The full path to the file this resource represents."""

    source = FullPath()
    """ The full path to a file to copy to target and patch """

    patch = File()

    strip = Integer(default=0)
    """ Strip the smallest prefix containing ``strip`` leading slashes from
    each file name found in the patch. """

    owner = String(default="root")
    """A unix username or UID who will own created objects. An owner that
    begins with a digit will be interpreted as a UID, otherwise it will be
    looked up using the python 'pwd' module."""

    group = String(default="root")
    """A unix group or GID who will own created objects. A group that begins
    with a digit will be interpreted as a GID, otherwise it will be looked up
    using the python 'grp' module."""

    mode = Octal(default=0o644)
    """A mode representation as an octal. This can begin with leading zeros if
    you like, but this is not required. DO NOT use yaml Octal representation
    (0o666), this will NOT work."""

    sensitive = Boolean(default=False)


class PatchApplyPolicy(Policy):

    """Apply a patch.

    You must provide a target.
    """

    resource = Patch
    name = "apply"
    default = True
    signature = (Present("name"),)



# Copyright 2011-2014 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Resources representing symbolic links. """

from fuselage.argument import FullPath, String
from fuselage.defaults import get_default_group, get_default_user
from fuselage.policy import Absent, Policy, Present
from fuselage.resource import Resource


class Link(Resource):

    """A resource representing a symbolic link. The link will be from `name`
    to `to`. If you specify owner, group and/or mode then these settings will
    be applied to the link itself, not to the object linked to.

    For example::

      Link:
        name: /etc/init.d/exampled
        to: /usr/local/example/sbin/exampled
        owner: root
        group: root

    """

    name = FullPath()
    """The name of the file this resource represents."""

    owner = String(default=lambda r: get_default_user())
    """A unix username or UID who will own created objects. An owner that
    begins with a digit will be interpreted as a UID, otherwise it will be
    looked up using the python 'pwd' module."""

    group = String(default=lambda r: get_default_group(r.owner))
    """A unix group or GID who will own created objects. A group that begins
    with a digit will be interpreted as a GID, otherwise it will be looked up
    using the python 'grp' module."""

    to = FullPath()
    """ The pathname to which to link the symlink. Dangling symlinks ARE
    considered errors in Yaybu. """


class LinkAppliedPolicy(Policy):
    resource = Link
    name = "apply"
    default = True
    signature = (
        Present("name"),
        Present("to"),
    )


class LinkRemovedPolicy(Policy):
    resource = Link
    name = "remove"
    default = False
    signature = (
        Present("name"),
        Absent("to"),
    )



# Copyright 2011-2014 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from fuselage.argument import Dict, FullPath, Integer, List, Octal, String
from fuselage.defaults import get_default_group, get_default_user
from fuselage.policy import XOR, Policy, Present
from fuselage.resource import Resource
from fuselage.utils import simple_str


class Execute(Resource):

    """Execute a command. This command is not executed in a shell - if you
    want a shell, run it (for example bash -c).

    For example::

        Execute:
          name: core_packages_apt_key
          command: apt-key adv --keyserver keyserver.ubuntu.com --recv-keys {{source.key}}

    A much more complex example. This shows executing a command if a checkout synchronises::

        Execute.foreach bi in {{flavour.base_images}}:
          name: base-image-{{bi}}
          policy:
              apply:
                  when: sync
                  on: /var/local/checkouts/ci
          command: ./vmbuilder-{{bi}}
          cwd: /var/local/checkouts/ci
          user: root

    """

    @property
    def implicit_id(self):
        implicit_id = super().implicit_id
        if not implicit_id:
            if self.command:
                implicit_id = simple_str(self.command)
            elif self.commands:
                implicit_id = simple_str("; ".join(self.commands))
        return implicit_id

    name = String()
    """ The name of this resource. This should be unique and descriptive, and
    is used so that resources can reference each other. """

    command = String()
    """ If you wish to run a single command, then this is the command. """

    commands = List()
    """ If you wish to run multiple commands, provide a list """

    cwd = FullPath(default="/")
    """ The current working directory in which to execute the command. """

    env = Dict()
    """

    The environment to provide to the command, for example::

        Execute:
            name: example
            command: echo $FOO
            environment:
                FOO: bar
    """

    returncode = Integer(default=0)
    """ The expected return code from the command, defaulting to 0. If the
    command does not return this return code then the resource is considered
    to be in error. """

    user = String(default=lambda r: get_default_user())
    """ The user to execute the command as.
    """

    group = String(default=lambda r: get_default_group(r.user))
    """ The group to execute the command as.
    """

    umask = Octal(default=0o022)
    """ The umask to use when executing this command """

    unless = String(default="")
    """ A command to run to determine is this execute should be actioned
    """

    creates = FullPath()
    """ The full path to a file that execution of this command creates. This
    is used like a "touch test" in a Makefile. If this file exists then the
    execute command will NOT be executed. """

    touch = FullPath()
    """ The full path to a file that yaybu will touch once this command has
    completed successfully. This is used like a "touch test" in a Makefile. If
    this file exists then the execute command will NOT be executed. """


class ExecutePolicy(Policy):

    """Execute the the command or commands provided.

    If user or group attributes are provided the command will be run using sudo."""

    resource = Execute
    name = "execute"
    default = True
    signature = (XOR(Present("command"), Present("commands")),)



# Copyright 2011-2014 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from fuselage import error, platform, provider, resources
from fuselage.changes import EnsureDirectory, ShellCommand


def dirname(p):
    i = p.rfind("/") + 1
    head = p[:i]
    if head and head != "/" * len(head):
        head = head.rstrip("/")
    return head


class Directory(provider.Provider):

    policies = (resources.directory.DirectoryAppliedPolicy,)

    def check_path(self, directory):
        if platform.isdir(directory):
            return

        frags = directory.split("/")
        path = "/"
        for i in frags:
            path = os.path.join(path, i)
            if not platform.exists(path):
                if self.resource.parents:
                    return
                self.raise_or_log(error.PathComponentMissing(path))
            if not platform.isdir(path):
                raise error.PathComponentNotDirectory(path)

    def apply(self):
        name = self.resource.name

        # FIXME: This should use os.path.name - but then what about tests that
        # rely on posix behaviour?
        # Need to mark those as skipped on nt and implement tests that exercise
        # simple behaviour on posix and nt
        self.check_path(dirname(name))

        return self.change(
            EnsureDirectory(
                name,
                self.resource.owner,
                self.resource.group,
                self.resource.mode,
                self.resource.parents,
            )
        )


class RemoveDirectory(provider.Provider):

    policies = (resources.directory.DirectoryRemovedPolicy,)

    def apply(self):
        name = self.resource.name

        if platform.exists(name) and not platform.isdir(name):
            raise error.InvalidProviderError(
                f"{self!r}: {name} exists and is not a directory"
            )
        if platform.exists(name):
            self.change(ShellCommand(["/bin/rmdir", self.resource.name]))
            changed = True
        else:
            changed = False
        return changed


class RemoveDirectoryRecursive(provider.Provider):

    policies = (resources.directory.DirectoryRemovedRecursivePolicy,)

    def apply(self):
        name = self.resource.name

        if platform.exists(name) and not platform.isdir(name):
            raise error.InvalidProviderError(
                f"{self!r}: {name} exists and is not a directory"
            )
        if platform.exists(name):
            self.change(ShellCommand(["/bin/rm", "-rf", self.resource.name]))
            changed = True
        else:
            changed = False
        return changed



# Copyright 2011-2014 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from fuselage import error, platform, provider, resources
from fuselage.changes import ShellCommand
from fuselage.utils import force_str


class Mount(provider.Provider):

    policies = (resources.mount.MountPolicy,)

    def check_path(self, directory):
        if platform.isdir(directory):
            return

        frags = directory.split("/")
        path = "/"
        for i in frags:
            path = os.path.join(path, i)
            if not platform.exists(path):
                if self.resource.parents:
                    return
                self.raise_or_log(error.PatchComponentMissing(path))
            if not platform.isdir(path):
                raise error.PathComponentNotDirectory(path)

    def get_all_active_mounts(self):
        path = force_str(platform.get("/proc/mounts"))
        d = {}
        for line in path.split("\n"):
            if not line.strip():
                continue

            split = line.split()
            d[split[1]] = {
                "fs_type": split[0],
                "mountpoint": split[1],
                "device": split[2],
                "options": split[3],
                "dump": split[4],
                "fsck": split[5],
            }
        return d

    def get_mount(self, path):
        return self.get_all_active_mounts()[path]

    def apply(self):
        name = self.resource.name

        self.check_path(name)

        try:
            self.get_mount(name)
            return

        except KeyError:
            command = ["mount"]

            fs_type = self.resource.fs_type
            if fs_type:
                if fs_type == "bind":
                    command.append("--bind")
                else:
                    command.extend(("-t", fs_type))
            command.append(self.resource.device)
            command.append(self.resource.name)

            options = self.resource.options
            if options:
                command.extend(("-o", options))

            self.change(
                ShellCommand(
                    command=command,
                )
            )
            return True



# Copyright 2011-2014 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from fuselage import error, platform, provider, resources
from fuselage.changes import ShellCommand


class Group(provider.Provider):

    policies = (resources.group.GroupApplyPolicy,)

    def get_group_info(self):
        fields = (
            "name",
            "passwd",
            "gid",
            "members",
        )

        try:
            info_tuple = platform.getgrnam(self.resource.name.encode("utf-8"))
        except KeyError:
            info = {f: None for f in fields}
            info["exists"] = False
            return info

        info = {"exists": True}
        for i, field in enumerate(fields):
            info[field] = info_tuple[i]

        return info

    def apply(self):
        changed = False
        info = self.get_group_info()

        if info["exists"]:
            command = ["groupmod"]
        else:
            command = ["groupadd"]
            changed = True

        gid = self.resource.gid
        if gid and info["gid"] != gid:
            command.extend(["--gid", str(self.resource.gid)])

        command.extend([self.resource.name])

        if not changed:
            return False

        try:
            self.change(ShellCommand(command))
        except error.SystemError as exc:
            raise error.InvalidGroup(
                "%s on %s failed with return code %d"
                % (command[0], self.resource, exc.returncode)
            )

        return True


class GroupRemove(provider.Provider):

    policies = (resources.group.GroupRemovePolicy,)

    def apply(self):
        try:
            platform.getgrnam(self.resource.name)
        except KeyError:
            # If we get a key errror then there is no such group. This is good.
            return False

        command = ["groupdel", self.resource.name]

        try:
            self.change(ShellCommand(command))
        except error.SystemError as exc:
            raise error.InvalidGroup(
                "groupdel on %s failed with return code %d"
                % (self.resource, exc.returncode)
            )

        return True



# Copyright 2013-2014 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from fuselage import platform
from fuselage.changes import base

from .attributes import AttributeChanger
from .execute import ShellCommand


class EnsureDirectory(base.Change):
    def __init__(self, path, owner, group, mode, recursive=False):
        self.path = path
        self.owner = owner
        self.group = group
        self.mode = mode
        self.recursive = recursive

    def apply(self, context):
        self.changed = False
        if not platform.exists(self.path):
            command = ["/bin/mkdir"]
            if self.recursive:
                command.append("-p")
            command.append(self.path)
            context.change(ShellCommand(command))
            self.changed = True

        ac = context.change(
            AttributeChanger(self.path, self.owner, self.group, self.mode)
        )

        self.changed = self.changed or ac.changed

        return self.changed
